"""
Module holds all stuff regarding Gatling tool usage

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import codecs
import json
import os
import re
import time
from collections import defaultdict

from bzt import TaurusConfigError, ToolError
from bzt.engine import ScenarioExecutor, Scenario
from bzt.modules.aggregator import ConsolidatingAggregator, ResultsReader
from bzt.modules.console import ExecutorWidget
from bzt.requests_model import HTTPRequest, SetVariables, HierarchicRequestParser
from bzt.utils import TclLibrary, EXE_SUFFIX, dehumanize_time, get_full_path, FileReader, RESOURCES_DIR, BetterDict
from bzt.utils import simple_body_dict, CALL_PROBLEMS, numeric_types
from bzt.utils import unzip, RequiredTool, JavaVM, shutdown_process, ensure_is_dict, is_windows


class GatlingScriptBuilder(object):
    def __init__(self, load, scenario, parent_logger, class_name, gatling_version=None):
        super(GatlingScriptBuilder, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.load = load
        self.feeder_names = {}
        self.scenario = scenario
        self.class_name = class_name
        if gatling_version is None:
            self.gatling_version = Gatling.VERSION
        else:
            self.gatling_version = gatling_version

    # add prefix 'http://' if user forgot it
    @staticmethod
    def fixed_addr(addr):
        if len(addr) > 0 and not addr.startswith('http'):
            return 'http://' + addr
        else:
            return addr

    @staticmethod
    def indent(text, level):
        return "  " * level + text

    def _get_http(self):
        default_address = self.scenario.get("default-address", "")

        http_str = '("%(addr)s")\n' % {'addr': self.fixed_addr(default_address)}

        if self.scenario.get("retrieve-resources", False):
            regex = self.scenario.get("retrieve-resources-regex")
            params = 'BlackList(), WhiteList("""%s""")' % regex if regex else ""
            http_str += self.indent(".inferHtmlResources(%s)\n" % params, level=2)

        if not self.scenario.get('store-cache', True):
            http_str += self.indent('.disableCaching\n', level=2)

        scenario_headers = self.scenario.get_headers()
        for key in scenario_headers:
            http_str += self.indent('.header("%(key)s", "%(val)s")\n' % {'key': key, 'val': scenario_headers[key]},
                                    level=2)
        return http_str

    def _get_exec(self):
        exec_str = ''
        for req in self.scenario.get_requests(parser=HierarchicRequestParser):
            if isinstance(req, SetVariables):
                if len(exec_str) > 0:
                    exec_str += '.'

                exec_str += "exec(\n"
                exec_str += self.indent("_", level=2)
                for k, v in sorted(req.mapping.items()):
                    exec_str += '.set("%s", "%s")' % (k.replace("\"", "\\\""), v.replace("\"", "\\\""))

                exec_str += "\n" + self.indent(")", level=1)
                continue

            if not isinstance(req, HTTPRequest):
                msg = "Gatling simulation generator doesn't support '%s' blocks, skipping"
                self.log.warning(msg, req.NAME)
                continue

            if len(exec_str) > 0:
                exec_str += '.'

            default_address = self.scenario.get("default-address")
            if default_address:
                url = req.url
            else:
                url = self.fixed_addr(req.url)

            exec_str += 'exec(\n'
            exec_template = self.indent('http("%(req_label)s").%(method)s("%(url)s")\n', level=2)
            exec_str += exec_template % {'req_label': req.label, 'method': req.method.lower(), 'url': url}

            for key in req.headers:
                exec_template = self.indent('.header("%(key)s", "%(val)s")\n', level=3)
                exec_str += exec_template % {'key': key, 'val': req.headers[key]}

            # todo: join with the same in get_sampler_pair
            if isinstance(req.body, (dict, list, numeric_types)):
                if req.get_header('content-type') == 'application/json' or isinstance(req.body, numeric_types):
                    req.body = json.dumps(req.body)
                elif not simple_body_dict(req.body):
                    self.log.debug('Header "Content-Type: application/json" is required for body: "%s"', req.body)
                    req.body = json.dumps(req.body)

            if isinstance(req.body, str):
                stmt = '.body(%(method)s("""%(body)s"""))\n' % {'method': 'StringBody', 'body': req.body}
                exec_str += self.indent(stmt, level=3)
            elif isinstance(req.body, dict):
                for key in sorted(req.body.keys()):
                    stmt = '.formParam("%(key)s", "%(val)s")\n' % {'key': key, 'val': req.body[key]}
                    exec_str += self.indent(stmt, level=3)
            elif req.body is not None:
                self.log.warning("Unknown body type: %s", req.body)

            exec_str += self.__get_assertions(req.config.get('assert', []))

            if not req.priority_option('follow-redirects', default=True):
                exec_str += self.indent('.disableFollowRedirect\n', level=3)

            exec_str += self.indent(')', level=1)

            think_time = int(dehumanize_time(req.get_think_time()))
            if think_time:
                exec_str += '.pause(%(think_time)s)' % {'think_time': think_time}

        return exec_str

    @staticmethod
    def __get_check_template(assertion):
        a_not = assertion.get('not', False)
        a_regexp = assertion.get('regexp', False)
        a_subject = assertion.get('subject', Scenario.FIELD_BODY)

        if a_subject == Scenario.FIELD_RESP_CODE:
            if a_not:
                res = 'status.not(%(sample)s)'
            else:
                res = 'status.is(%(sample)s)'
        elif a_subject == Scenario.FIELD_HEADERS:
            res = ''
        else:  # FIELD_BODY
            if a_regexp:
                res = 'regex("""%(sample)s""").'
            else:
                res = 'substring("""%(sample)s""").'
            if a_not:
                res += 'notExists'
            else:
                res += 'exists'
        return res

    def __get_assertions(self, assertions):
        if len(assertions) == 0:
            return ''

        first_check = True
        check_result = self.indent('.check(\n', level=3)

        for idx, assertion in enumerate(assertions):
            assertion = ensure_is_dict(assertions, idx, "contains")

            error_str = 'You must specify "contains" parameter for assertion item'
            a_contains = assertion.get('contains', TaurusConfigError(error_str))

            check_template = self.__get_check_template(assertion)

            if check_template == '':  # FIELD_HEADERS
                self.log.warning('Sorry, but "headers" subject is not implemented for gatling asserts')
                return ''

            if not isinstance(a_contains, list):
                a_contains = [a_contains]
            for sample in a_contains:
                if not first_check:
                    check_result += ',\n'
                check_result += self.indent(check_template % {'sample': sample}, level=4)
                first_check = False

        check_result += '\n' + self.indent(')', level=3) + '\n'

        return check_result

    def _get_feeder_name(self, source_filename):
        base_feeder_name = ".".join(os.path.basename(source_filename).split(".")[:-1])
        base_feeder_name = re.sub(r'[^A-Za-z0-9_]', '', base_feeder_name) + "Feed"

        index = 0
        feeder_name = base_feeder_name
        while feeder_name in self.feeder_names and self.feeder_names[feeder_name] != source_filename:
            index += 1
            feeder_name = base_feeder_name + "_%s" % index

        if feeder_name not in self.feeder_names:
            self.feeder_names[feeder_name] = source_filename

        return feeder_name

    def _get_feeders(self):
        feeders_def = ""
        feeding = ""

        for source in self.scenario.get_data_sources():
            path = self.scenario.engine.find_file(source["path"])

            delimiter = source.get('delimiter', None)
            loop_over = source.get("loop", True)
            var_name = self._get_feeder_name(path)
            params = dict(varname=var_name, filename=path, delimiter=delimiter)
            if delimiter is not None:
                tpl = """val %(varname)s = separatedValues("%(filename)s", '%(delimiter)s')"""
            else:
                tpl = 'val %(varname)s = csv("%(filename)s")'
            line = self.indent(tpl % params, level=1)
            if loop_over:
                line += '.circular'
            feeders_def += line + '\n'
            feeding += "feed(%s)." % var_name

        if feeders_def:
            feeders_def = '\n' + feeders_def

        return feeders_def, feeding

    def gen_test_case(self):
        template_path = os.path.join(RESOURCES_DIR, "gatling", "v3_script.tpl")

        with open(template_path) as template_file:
            template_line = template_file.read()

        feeders_def, feeding = self._get_feeders()

        params = {
            'class_name': self.class_name,
            'httpConf': self._get_http(),
            '_exec': self._get_exec(),
            'feeders': feeders_def,
            'feeding': feeding,
        }
        return template_line % params


class GatlingExecutor(ScenarioExecutor):
    """
    Gatling executor module
    """

    def __init__(self):
        super(GatlingExecutor, self).__init__()
        self.script = None
        self.process = None
        self.end_time = None
        self.retcode = None
        self.simulation_started = False
        self.dir_prefix = "gatling-%s" % id(self)
        self.tool = None

    def get_cp_from_files(self):
        jar_files = []
        files = self.execution.get('files', [])
        for candidate in files:
            candidate = self.engine.find_file(candidate)
            if os.path.isfile(candidate) and candidate.lower().endswith('.jar'):
                jar_files.append(candidate)
            elif os.path.isdir(candidate):
                for element in os.listdir(candidate):
                    element = os.path.join(candidate, element)
                    if os.path.isfile(element) and element.lower().endswith('.jar'):
                        jar_files.append(element)

        return jar_files

    def get_additional_classpath(self):
        cp = self.get_scenario().get("additional-classpath", [])
        cp.extend(self.settings.get("additional-classpath", []))
        return cp

    def prepare(self):
        super(GatlingExecutor, self).prepare()

        self.install_required_tools()
        scenario = self.get_scenario()

        self.env.set({"GATLING_HOME": self.tool.tool_dir})

        cpath = self.get_additional_classpath()
        self.log.debug("Classpath for Gatling: %s", cpath)

        for element in cpath:
            self.env.add_path({"JAVA_CLASSPATH": element})
            self.env.add_path({"COMPILATION_CLASSPATH": element})

        new_name = self.engine.create_artifact('gatling-launcher', EXE_SUFFIX)
        self.log.debug("Building Gatling launcher: %s", new_name)
        self.tool.build_launcher(new_name)

        self.script = self.get_script_path()
        if not self.script:
            if "requests" in scenario:
                self.get_scenario()['simulation'], self.script = self.__generate_script()
            else:
                msg = "There must be a script file or requests for its generation "
                msg += "to run Gatling tool (%s)" % self.execution.get('scenario')
                raise TaurusConfigError(msg)

        self.dir_prefix = self.settings.get("dir-prefix", self.dir_prefix)

        self.stdout = open(self.engine.create_artifact("gatling", ".out"), "w")
        self.stderr = open(self.engine.create_artifact("gatling", ".err"), "w")

        self.reader = DataLogReader(self.engine.artifacts_dir, self.log, self.dir_prefix)
        if isinstance(self.engine.aggregator, ConsolidatingAggregator):
            self.engine.aggregator.add_underling(self.reader)

    def __generate_script(self):
        simulation = "TaurusSimulation_%s" % id(self)
        file_name = self.engine.create_artifact(simulation, ".scala")
        gen_script = GatlingScriptBuilder(self.get_load(), self.get_scenario(), self.log, simulation, self.tool.version)
        with codecs.open(file_name, 'w', encoding='utf-8') as script:
            script.write(gen_script.gen_test_case())

        return simulation, file_name

    def _get_simulation_props(self):
        props = {}
        if os.path.isfile(self.script):
            if self.script.endswith('.jar'):
                self.env.add_path({"JAVA_CLASSPATH": self.script})
                self.env.add_path({"COMPILATION_CLASSPATH": self.script})
            else:
                props['gatling.core.directory.simulations'] = get_full_path(self.script, step_up=1)
        else:
            props['gatling.core.directory.simulations'] = self.script

        simulation = self.get_scenario().get("simulation")
        if simulation:
            props['gatling.core.simulationClass'] = simulation
        else:
            props['gatling.core.runDescription'] = "Taurus_Test"
        return props

    def _get_load_props(self):
        load = self.get_load()
        props = {}
        if load.concurrency:
            props['concurrency'] = load.concurrency
        if load.ramp_up is not None:
            props['ramp-up'] = int(load.ramp_up)
        if load.hold is not None:
            props['hold-for'] = int(load.hold)
        if load.iterations:
            props['iterations'] = int(load.iterations)
        if load.throughput:
            if load.duration:
                props['throughput'] = load.throughput
            else:
                self.log.warning("You should set up 'ramp-up' and/or 'hold-for' for usage of 'throughput'")
        return props

    def _get_scenario_props(self):
        props = {}
        scenario = self.get_scenario()
        timeout = scenario.get('timeout', None)
        if timeout is not None:
            props['gatling.http.ahc.requestTimeout'] = int(dehumanize_time(timeout) * 1000)
        return props

    def _set_env(self):
        props = BetterDict()
        props.merge(self.settings.get('properties'))
        props.merge(self.get_scenario().get("properties"))

        props['gatling.core.outputDirectoryBaseName'] = self.dir_prefix
        props['gatling.core.directory.resources'] = self.engine.artifacts_dir
        props['gatling.core.directory.results'] = self.engine.artifacts_dir

        props.merge(self._get_simulation_props())
        props.merge(self._get_load_props())
        props.merge(self._get_scenario_props())
        for key in sorted(props.keys()):
            prop = props[key]
            val_tpl = "%s"

            if isinstance(prop, str):
                if not is_windows():  # extend properties support (contained separators/quotes/etc.) on lin/mac
                    val_tpl = "%r"

            self.env.add_java_param({"JAVA_OPTS": ("-D%s=" + val_tpl) % (key, prop)})

        self.env.set({"NO_PAUSE": "TRUE"})
        self.env.add_java_param({"JAVA_OPTS": self.settings.get("java-opts", None)})

        self.log.debug('JAVA_OPTS: "%s"', self.env.get("JAVA_OPTS"))

    def startup(self):
        self._set_env()
        self.process = self._execute([self.tool.tool_path])

    def check(self):
        """
        Checks if tool is still running. Also checks if resulting logs contains
        any data and throws exception otherwise.

        :return: bool
        :raise TaurusConfigError:
        :raise TaurusToolError:
        """
        self.retcode = self.process.poll()

        # detect interactive mode and raise exception if it found
        if not self.simulation_started:
            wrong_line = "Choose a simulation number:"
            with open(self.stdout.name) as out:
                file_header = out.read(1024)
            if wrong_line in file_header:  # gatling can't select test scenario
                scenarios = file_header[file_header.find(wrong_line) + len(wrong_line):].rstrip()
                msg = 'Several gatling simulations are found, you must '
                msg += 'specify one of them to use in "simulation" option: %s' % scenarios
                raise TaurusConfigError(msg)
            if 'started...' in file_header:
                self.simulation_started = True

        if self.retcode is None:
            return False
        elif self.retcode == 0:
            return True
        else:
            raise ToolError("Gatling tool exited with non-zero code: %s" % self.retcode, self.get_error_diagnostics())

    def shutdown(self):
        """
        If tool is still running - let's stop it.
        """
        shutdown_process(self.process, self.log)

        if self.start_time:
            self.end_time = time.time()
            self.log.debug("Gatling worked for %s seconds", self.end_time - self.start_time)

    def post_process(self):
        """
        Save data log as artifact
        """
        if self.reader and self.reader.file and self.reader.file.name:
            self.engine.existing_artifact(self.reader.file.name)
        super(GatlingExecutor, self).post_process()

    def install_required_tools(self):
        self.tool = self._get_tool(Gatling, config=self.settings)
        java = self._get_tool(JavaVM)
        required_tools = [self._get_tool(TclLibrary), java, self.tool]

        for tool in required_tools:
            if not tool.check_if_installed():
                tool.install()

    def get_widget(self):
        if not self.widget:
            simulation = self.get_scenario().get('simulation', None)
            if simulation == "TaurusSimulation_%s" % id(self):
                simulation = 'generated script'
            if simulation is None:
                simulation = os.path.basename(self.script)
            self.widget = ExecutorWidget(self, 'Gatling: %s' % simulation)
        return self.widget

    def resource_files(self):
        files = []
        script = self.get_script_path()
        if script:
            files.append(script)
        else:
            for source in self.get_scenario().get_data_sources():
                source_path = self.engine.find_file(source["path"])
                files.append(source_path)
        files.extend(self.get_additional_classpath())
        return files

    def get_error_diagnostics(self):
        diagnostics = []
        if self.stdout is not None:
            with open(self.stdout.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Gatling STDOUT:\n" + contents)
        if self.stderr is not None:
            with open(self.stderr.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Gatling STDERR:\n" + contents)
        if self.reader and self.reader.file and self.reader.file.name:
            with open(self.reader.file.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Simulation log:\n" + contents)
        return diagnostics


class DataLogReader(ResultsReader):
    """ Class to read KPI from data log """

    def __init__(self, basedir, parent_logger, dir_prefix):
        super(DataLogReader, self).__init__()
        self.concurrency = 0
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.basedir = basedir
        self.file = FileReader(file_opener=self.open_fds, parent_logger=self.log)
        self.partial_buffer = ""
        self.delimiter = "\t"
        self.dir_prefix = dir_prefix
        self.guessed_gatling_version = None
        self._group_errors = defaultdict(set)

    def _extract_log(self, fields):
        """
        Extract stats from Gatling format of version 3.1 and after
        :param fields:
        :return:
        """
        # 0 ${RequestRecordHeader.value}
        # 1 $scenario
        # -|2 $userId, absent in Gatling 3.4+
        # 2 ${serializeGroups(groupHierarchy)}
        # 3 $label
        # 4 $startTimestamp
        # 5 $endTimestamp
        # 6 $status
        # [7] ${serializeMessage(message)}${serializeExtraInfo(extraInfo)}

        if fields[0].strip() == "USER":
            if self.guessed_gatling_version < "3.4+":
                del fields[2]  # ignore obsolete $userId
            if fields[2].strip() == "START":
                self.concurrency += 1
            elif fields[2].strip() == "END":
                self.concurrency -= 1

        elif fields[0].strip() == "GROUP":
            del fields[0]
            return self.__parse_group(fields)
        elif fields[0].strip() == "REQUEST":
            del fields[0]
            return self.__parse_request(fields)
        else:
            return None

    def __parse_group(self, fields):
        latency = 0.0
        con_time = 0.0

        if len(fields) < 3:
            label = ""
            t_stamp = int(fields[1]) / 1000.0
            r_time = 0
            error = fields[0]
            r_code = "N/A"
        else:
            if self.guessed_gatling_version < "3.4+":
                del fields[0]  # ignore obsolete $userId
            label = fields[0]
            if ',' in label:
                return None  # skip nested groups for now
            t_stamp = int(fields[2]) / 1000.0
            r_time = int(fields[3]) / 1000.0

            if label in self._group_errors:
                error = ';'.join(self._group_errors.pop(label))
            else:
                error = None

            if fields[4] == 'OK':
                r_code = '200'
            else:
                r_code = self.__rc_from_msg(fields[-1])

        return int(t_stamp), label, r_time, con_time, latency, r_code, error

    def __parse_request(self, fields):
        # see LogFileDataWriter.ResponseMessageSerializer in gatling-core

        if self.guessed_gatling_version < "3.4+":
            del fields[0]  # ignore obsolete $userId

        if len(fields) >= 6 and fields[5]:
            error = fields[5]
        else:
            error = None

        req_hierarchy = fields[0].split(',')[0]
        if req_hierarchy:
            if error:
                self._group_errors[req_hierarchy].add(error)
            return None

        label = fields[1]
        t_stamp = int(fields[3]) / 1000.0
        r_time = (int(fields[3]) - int(fields[2])) / 1000.0
        latency = 0.0
        con_time = 0.0
        if fields[4] == 'OK':
            r_code = '200'
        else:
            r_code = self.__rc_from_msg(fields[-1])

        return int(t_stamp), label, r_time, con_time, latency, r_code, error

    def __rc_from_msg(self, msg):
        _tmp_rc = msg.split("but actually ")[-1]  # gatling-core/src/main/scala/io/gatling/core/check/Validator.scala

        if _tmp_rc.startswith("unexpectedly "):
            _tmp_rc = _tmp_rc[len("unexpectedly "):]
        if _tmp_rc.startswith("found "):
            _tmp_rc = _tmp_rc[len("found "):]

        parts = _tmp_rc.split(' ')
        if len(parts) > 1 and parts[1] == 'is':
            _tmp_rc = parts[0]

        return _tmp_rc if _tmp_rc.isdigit() else 'N/A'

    def _guess_gatling_version(self, fields):
        if fields and fields[-1].strip() < "3.4":
            return "3.3.X"
        elif fields[-1].strip() >= "3.4":
            return "3.4+"
        else:
            return ""

    def _extract_log_data(self, fields):
        if self.guessed_gatling_version is None:
            self.guessed_gatling_version = self._guess_gatling_version(fields)

        return self._extract_log(fields) if self.guessed_gatling_version else None

    def _read(self, last_pass=False):
        """
        Generator method that returns next portion of data

        :param last_pass:
        """
        lines = self.file.get_lines(size=1024 * 1024, last_pass=last_pass)

        for line in lines:
            if not line.endswith("\n"):
                self.partial_buffer += line
                continue

            line = "%s%s" % (self.partial_buffer, line)
            self.partial_buffer = ""

            line = line.strip()
            fields = line.split(self.delimiter)

            data = self._extract_log_data(fields)
            if data is None:
                continue

            t_stamp, label, r_time, con_time, latency, r_code, error = data
            bytes_count = None
            yield t_stamp, label, self.concurrency, r_time, con_time, latency, r_code, error, '', bytes_count

    def open_fds(self, filename):
        """
        open gatling simulation.log
        """
        if os.path.isdir(self.basedir):
            prog = re.compile("^%s-[0-9]+$" % self.dir_prefix)

            for fname in os.listdir(self.basedir):
                if prog.match(fname):
                    filename = os.path.join(self.basedir, fname, "simulation.log")
                    break

            if not filename or not os.path.isfile(filename):
                self.log.debug('simulation.log not found')
                return
        elif os.path.isfile(self.basedir):
            filename = self.basedir
        else:
            self.log.debug('Path not found: %s', self.basedir)
            return

        if not os.path.getsize(filename):
            self.log.debug('simulation.log is empty')
        else:
            return open(filename, 'rb')


class Gatling(RequiredTool):
    """
    Gatling tool
    """
    DOWNLOAD_LINK = "https://repo1.maven.org/maven2/io/gatling/highcharts/gatling-charts-highcharts-bundle" \
                    "/{version}/gatling-charts-highcharts-bundle-{version}-bundle.zip"
    VERSION = "3.5.1"
    LOCAL_PATH = "~/.bzt/gatling-taurus/{version}/bin/gatling{suffix}"

    def __init__(self, config=None, **kwargs):
        settings = config or {}
        version = settings.get("version", self.VERSION)
        def_path = self.LOCAL_PATH.format(version=version, suffix=EXE_SUFFIX)
        gatling_path = get_full_path(settings.get("path", def_path))
        download_link = settings.get("download-link", self.DOWNLOAD_LINK).format(version=version)
        super(Gatling, self).__init__(tool_path=gatling_path, download_link=download_link, version=version, **kwargs)

        self.tool_dir = get_full_path(self.tool_path, step_up=2)

    def check_if_installed(self):
        self.log.debug("Trying Gatling...")
        try:
            out, err = self.call([self.tool_path, '--help'])
            self.log.debug("Gatling check output: %s", out)
        except CALL_PROBLEMS as exc:
            self.log.info("Gatling check failed: %s", exc)
            return False

        if err:
            self.log.warning("Gatling check stderr: %s", err)
        return True

    def install(self):
        dest = get_full_path(self.tool_path, step_up=2)
        self.log.info("Will install %s into %s", self.tool_name, dest)
        gatling_dist = self._download(use_link=True)
        self.log.info("Unzipping %s", gatling_dist)
        unzip(gatling_dist, dest, 'gatling-charts-highcharts-bundle-' + self.version)
        os.remove(gatling_dist)
        os.chmod(get_full_path(self.tool_path), 0o755)
        self.log.info("Installed Gatling successfully")
        if not self.check_if_installed():
            raise ToolError("Unable to run %s after installation!" % self.tool_name)

    def build_launcher(self, new_name):  # legacy, for v2 only
        modified_lines = []
        mod_success = False

        with open(self.tool_path) as fds:
            for line in fds.readlines():
                if is_windows():
                    if line.startswith('set COMPILER_CLASSPATH='):
                        mod_success = True
                        line = line.rstrip() + ';%COMPILATION_CLASSPATH%\n'  # add from env
                    elif line.startswith('set GATLING_CLASSPATH='):
                        mod_success = True
                        line = line.rstrip() + ';%JAVA_CLASSPATH%\n'  # add from env
                else:
                    if line.startswith('COMPILER_CLASSPATH='):
                        mod_success = True
                        line = line.rstrip()[:-1] + '${COMPILATION_CLASSPATH}"\n'  # add from env
                    elif line.startswith('GATLING_CLASSPATH='):
                        mod_success = True
                        line = line.rstrip()[:-1] + '${JAVA_CLASSPATH}"\n'  # add from env
                    elif line.startswith('"$JAVA"'):
                        line = 'eval ' + line
                modified_lines.append(line)

        if not mod_success:
            raise ToolError("Can't modify gatling launcher for jar usage, ability isn't supported")

        self.tool_path = new_name

        with open(self.tool_path, 'w') as modified:
            modified.writelines(modified_lines)

        if not is_windows():
            os.chmod(self.tool_path, 0o755)



import json
import logging
from lxml import etree

from bzt.jmx.base import JMX
from bzt.jmx.tools import ProtocolHandler
from bzt.utils import numeric_types, get_host_ips, simple_body_dict

LOG = logging.getLogger("")


class HTTPProtocolHandler(ProtocolHandler):
    def get_toplevel_elements(self, scenario):
        return self._gen_managers(scenario) + self._gen_defaults(scenario)

    def _gen_managers(self, scenario):
        elements = []
        if scenario.get("store-cache", True):
            elements.append(JMX._get_cache_mgr())
            elements.append(etree.Element("hashTree"))
        if scenario.get("store-cookie", True):
            elements.append(JMX._get_cookie_mgr(scenario))
            elements.append(etree.Element("hashTree"))
        if scenario.get("use-dns-cache-mgr", True):
            elements.append(JMX.get_dns_cache_mgr())
            elements.append(etree.Element("hashTree"))
            self.system_props.merge({"system-properties": {"sun.net.inetaddr.ttl": 0}})
        return elements

    def _gen_defaults(self, scenario):
        default_address = scenario.get("default-address", None)
        retrieve_resources = scenario.get("retrieve-resources", True)
        resources_regex = scenario.get("retrieve-resources-regex", None)
        concurrent_pool_size = scenario.get("concurrent-pool-size", 4)

        content_encoding = scenario.get("content-encoding", None)

        timeout = scenario.get("timeout", "30s")
        timeout = self.safe_time(timeout)
        elements = [JMX._get_http_defaults(default_address, timeout, retrieve_resources,
                                           concurrent_pool_size, content_encoding, resources_regex),
                    etree.Element("hashTree")]
        return elements

    def get_sampler_pair(self, request):
        # convert body to string
        if isinstance(request.body, (dict, list, numeric_types)):
            if request.get_header('content-type') == 'application/json' or isinstance(request.body, numeric_types):
                request.body = json.dumps(request.body)
            elif not simple_body_dict(request.body):
                LOG.debug('Header "Content-Type: application/json" is required for body: "%s"', request.body)
                request.body = json.dumps(request.body)

        use_random_host_ip = request.priority_option('random-source-ip', default=False)
        host_ips = get_host_ips(filter_loopbacks=True) if use_random_host_ip else []

        files = request.upload_files
        body_file = request.config.get("body-file")
        has_file_for_body = not (request.body or files) and body_file

        # method can be put, post, or even variable
        if has_file_for_body and request.method != "GET":
            files = [{"path": body_file}]

        timeout = self.safe_time(request.config.get("timeout"))

        http = JMX._get_http_request(request.url, request.label, request.method, timeout, request.body,
                                     request.priority_option('keepalive', default=True),
                                     files, request.content_encoding,
                                     request.priority_option('follow-redirects', default=True),
                                     use_random_host_ip, host_ips)

        children = etree.Element("hashTree")

        if request.headers:
            children.append(JMX._get_header_mgr(request.headers))
            children.append(etree.Element("hashTree"))

        return http, children



"""
Copyright 2021 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import os
from subprocess import PIPE
from bzt import TaurusConfigError, ToolError
from bzt.modules import ScenarioExecutor
from bzt.modules.console import ExecutorWidget
from bzt.modules.aggregator import ResultsReader, ConsolidatingAggregator
from bzt.utils import RequiredTool, CALL_PROBLEMS, FileReader, shutdown_process, get_full_path, is_windows, is_mac, \
    untar


class VegetaExecutor(ScenarioExecutor):
    def __init__(self):
        super(VegetaExecutor, self).__init__()
        self.output_file = None
        self.log_file = None
        self.script = None
        self.process = None
        self.vegeta = None
        self.kpi_file = None
        self.scenario = None

    def prepare(self):
        super(VegetaExecutor, self).prepare()
        self.scenario = self.get_scenario()
        self.install_required_tools()

        self.script = self.get_script_path()
        if not self.script:
            requests = self.scenario.get_requests()
            if not requests:
                raise TaurusConfigError("Either 'script' or 'scenario' should be present for Vegeta executor")
            self.script = os.path.join(self.engine.artifacts_dir, "vegeta.txt")
            with open(self.script, "w") as f:
                i = 0
                for request in requests:
                    f.write("{} {}\n".format(request.method, request.url))
                    headers = "\n".join(["{}: {}".format(key, value) for key, value in request.headers.items()])
                    if headers:
                        f.write("{}\n".format(headers))
                    if request.body:
                        json_body_file = os.path.join(self.engine.artifacts_dir, "body-{}.json".format(i))
                        with open(json_body_file, "w") as g:
                            g.write(json.dumps(request.body))
                        f.write("@{}\n".format(json_body_file))
                    f.write("\n")
                    i += 1

        self.stdout = open(self.engine.create_artifact("Vegeta", ".out"), "w")
        self.stderr = open(self.engine.create_artifact("Vegeta", ".err"), "w")

        self.kpi_file = self.engine.create_artifact("kpi", ".csv")
        self.reader = VegetaLogReader(self.kpi_file, self.log)
        if isinstance(self.engine.aggregator, ConsolidatingAggregator):
            self.engine.aggregator.add_underling(self.reader)

    def startup(self):
        cmdline = [self.vegeta.tool_path, "attack", "-targets", self.script]
        load = self.get_load()

        if load.throughput:
            cmdline += ['-rate', str(load.throughput)]

        if load.hold:
            cmdline += ['-duration', str(int(load.hold)) + "s"]

        if load.concurrency:
            cmdline += ['-max-workers', str(int(load.concurrency))]

        if self.scenario and 'timeout' in self.scenario:
            cmdline += ['-timeout', str(int(self.scenario.get('timeout'))) + "s"]

        user_cmd = self.settings.get("cmdline")
        if user_cmd:
            cmdline += user_cmd.split(" ")

        self.process = self._execute(cmdline, stdout=PIPE, shell=False)
        with open(self.kpi_file, 'wb') as f:
            self._execute([self.vegeta.tool_path, "encode", "-to=csv"], stdin=self.process.stdout, stdout=f, shell=False)

    def get_widget(self):
        if not self.widget:
            label = "%s" % self
            self.widget = ExecutorWidget(self, "Vegeta: " + label.split('/')[1])
        return self.widget

    def check(self):
        retcode = self.process.poll()
        if retcode is not None:
            ToolError(f"Vegeta tool exited with non-zero code: {retcode}")
            return True
        return False

    def shutdown(self):
        shutdown_process(self.process, self.log)

    def post_process(self):
        if self.kpi_file:
            self.engine.existing_artifact(self.kpi_file)
        super(VegetaExecutor, self).post_process()

    def install_required_tools(self):
        self.vegeta = self._get_tool(Vegeta, config=self.settings)
        self.vegeta.tool_name = self.vegeta.tool_name.lower()
        if not self.vegeta.check_if_installed():
            self.vegeta.install()

    def resource_files(self):
        return [self.get_script_path(required=True)]


class VegetaLogReader(ResultsReader):
    def __init__(self, filename, parent_logger):
        super(VegetaLogReader, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.file = FileReader(filename=filename, parent_logger=self.log)

    def _read(self, last_pass=False):
        lines = self.file.get_lines(size=1024 * 1024, last_pass=last_pass)

        for line in lines:
            log_vals = [val.strip() for val in line.split(',')]

            _tstamp = int(log_vals[0][:10])
            _url = log_vals[10]
            _concur = 1
            _etime = float(log_vals[2]) / 1000000000.0
            _con_time = 0
            _latency = 0
            _rstatus = log_vals[1]
            _error = log_vals[5] or None
            _bytes = int(log_vals[4])

            yield _tstamp, _url, _concur, _etime, _con_time, _latency, _rstatus, _error, '', _bytes


class Vegeta(RequiredTool):
    DOWNLOAD_LINK = \
        "https://github.com/tsenart/vegeta/releases/download/v{version}/vegeta_{version}_{platform}_amd64.tar.gz "
    VERSION = "12.8.4"
    LOCAL_PATH = "~/.bzt/vegeta-taurus/{version}/"

    def __init__(self, config=None, **kwargs):
        settings = config or {}
        version = settings.get("version", self.VERSION)
        self.tool_path = get_full_path(settings.get("path", self.LOCAL_PATH.format(version=version) + 'vegeta'))
        if not is_windows():
            platform = 'darwin' if is_mac() else 'linux'
            download_link = settings.get("download-link", self.DOWNLOAD_LINK).format(version=version, platform=platform)
        else:
            download_link = ''
        super(Vegeta, self).__init__(tool_path=self.tool_path, download_link=download_link, version=version, **kwargs)

    def check_if_installed(self):
        self.log.debug('Checking Vegeta Framework: %s' % self.tool_path)
        try:
            out, err = self.call([self.tool_path, '-version'])
        except CALL_PROBLEMS as exc:
            self.log.warning("%s check failed: %s", self.tool_name, exc)
            return False

        if err:
            out += err
        self.log.debug("Vegeta output: %s", out)
        return True

    def install(self):
        if is_windows():
            raise ToolError("Unable to install Vegeta on Windows! Manual installation required.")

        dest = get_full_path(self.tool_path, step_up=1)
        if not os.path.exists(dest):
            os.makedirs(dest)

        self.log.info("Will install %s into %s", self.tool_name, dest)
        vegeta_dist = self._download(use_link=True)

        self.log.info("Untaring %s", vegeta_dist)
        untar(vegeta_dist, dest, rel_path='vegeta')
        os.remove(vegeta_dist)
        os.chmod(get_full_path(self.tool_path), 0o755)
        self.log.info("Installed Vegeta successfully")

        if not self.check_if_installed():
            raise ToolError("Unable to run %s after installation!" % self.tool_name)



"""
Copyright 2016 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from abc import abstractmethod

from bzt.engine import Aggregator
from bzt.modules.aggregator import ResultsReader
from bzt.utils import BetterDict, iteritems, LDJSONReader


class FunctionalAggregator(Aggregator):
    """
    :type listeners: list[FunctionalAggregatorListener]
    :type underlings: list[FunctionalResultsReader]
    :type cumulative_results: ResultsTree
    """

    def __init__(self):
        super(FunctionalAggregator, self).__init__(is_functional=True)
        self.underlings = []
        self.listeners = []
        self.cumulative_results = ResultsTree()

    def add_underling(self, reader):
        assert isinstance(reader, FunctionalResultsReader)
        self.underlings.append(reader)

    def add_listener(self, listener):
        assert isinstance(listener, FunctionalAggregatorListener)
        self.listeners.append(listener)

    def prepare(self):
        # TODO: setup, read settings
        pass

    def process_readers(self, last_pass=False):
        new_results = ResultsTree()

        for reader in self.underlings:
            for sample in reader.read(last_pass):
                new_results.add_sample(sample)

        if new_results:
            self.cumulative_results.merge(new_results)
            for listener in self.listeners:
                listener.aggregated_results(new_results, self.cumulative_results)

    def check(self):
        self.process_readers()
        return False

    def post_process(self):
        self.process_readers(last_pass=True)


class FunctionalSample(object):
    def __init__(
            self, test_case, test_suite, status,
            start_time, duration, error_msg, error_trace,
            extras=None, subsamples=None, path=None,
    ):
        # test_case: str - name of test case (method)
        # test_suite: str - name of test suite (class)
        # status: str - test status (PASSED / FAILED / BROKEN / SKIPPED)
        # start_time: float - epoch
        # duration: float - test duration (in seconds)
        # error_msg: str - one-line error message
        # error_trace: str - error stacktrace
        # extras: dict - additional test info (description, file, full_name)
        # subsamples: list - list of subsamples
        # path: list - list of path components: [{"type": str, "value": str}]
        self.test_case = test_case
        self.test_suite = test_suite
        self.status = status
        self.start_time = start_time
        self.duration = duration
        self.error_msg = error_msg
        self.error_trace = error_trace
        self.extras = extras or {}
        self.subsamples = subsamples or []
        self.path = path or []

    def get_fqn(self):
        if self.path:
            return '.'.join(comp["value"] for comp in self.path)
        else:
            return self.test_suite + '.' + self.test_case

    def get_short_name(self):
        if self.path:
            return '.'.join(comp["value"] for comp in self.path[-2:])
        else:
            return self.test_suite + '.' + self.test_case

    def get_type(self):
        if self.path:
            return self.path[-1]["type"]
        return None


class ResultsTree(BetterDict):
    def __init__(self):
        super(ResultsTree, self).__init__()

    def add_sample(self, sample):
        """
        :type sample: FunctionalSample
        """
        test_suite = sample.test_suite
        self.get(test_suite, [], force_set=True).append(sample)

    def test_suites(self):
        return [key for key, _ in iteritems(self)]

    def test_cases(self, suite_name):
        return self.get(suite_name, [])


class FunctionalResultsReader(object):
    @abstractmethod
    def read(self, last_pass=False):
        """Yields functional samples"""
        yield


class FunctionalAggregatorListener(object):
    @abstractmethod
    def aggregated_results(self, results, cumulative_results):
        """
        Callback that gets called every time aggregator processes new test results.
        :type results: ResultsTree
        :type cumulative_results: ResultsTree
        """
        pass


class TestReportReader(object):
    SAMPLE_KEYS = [
        "test_case",  # str
        "test_suite",  # str
        "status",  # str
        "start_time",  # float, epoch
        "duration",  # float, in seconds
        "error_msg",  # short string
        "error_trace",  # multiline string
        "extras",  # dict
        "subsamples",  # list of samples
        "assertions",  # list of dicts, {"name": str, "failed": bool, "error_msg": str, "error_trace": str}
        "path"  # list of components, [{"value": "test_Something", "type": "module"},
        #                      {"value": "TestAPI", "type": "class"},
        #                      {"value": "test_heartbeat", "type": "method"}
        #                      {"value": "index page": "type": "transaction"}
        #                      {"value": "http://blazedemo.com/": "type": "request"}
    ]
    TEST_STATUSES = ("PASSED", "FAILED", "BROKEN", "SKIPPED")
    FAILING_TESTS_STATUSES = ("FAILED", "BROKEN")

    def __init__(self, filename, parent_logger):
        super(TestReportReader, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.json_reader = LDJSONReader(filename, self.log)

    @staticmethod
    def process_label(label):
        if isinstance(label, str):
            parts = label.split('_', 2)  # 'test_01_feeling_good'
            if len(parts) == 3 and parts[0] == 'test' and parts[1].isdigit():
                return parts[2]
        return label

    def process_path(self, path):
        if isinstance(path, dict):
            test_suite = ".".join(part["value"] for part in path[:-1])
            test_case = path[-1]["value"]
            return test_suite, test_case
        return None

    def read(self, last_pass=False):
        for row in self.json_reader.read(last_pass):
            if "path" in row:
                processed_path = self.process_path(row["path"])
                if processed_path is not None:
                    row["test_suite"], row["test_case"] = processed_path
            row["test_case"] = self.process_label(row["test_case"])
            yield row


class LoadSamplesReader(ResultsReader):
    STATUS_TO_CODE = {
        "PASSED": "200",
        "SKIPPED": "300",
        "FAILED": "400",
        "BROKEN": "500",
    }

    def __init__(self, filename, parent_logger):
        super(LoadSamplesReader, self).__init__()
        self.report_reader = TestReportReader(filename, parent_logger)
        self.read_records = 0

    def extract_sample(self, item):
        if item["status"] == 'SKIPPED':
            return  # we ignore skipped samples to not skew results

        tstmp = int(item["start_time"])
        label = item["test_case"]
        concur = 1
        rtm = item["duration"]
        cnn = 0
        ltc = 0
        rcd = self.STATUS_TO_CODE.get(item["status"], "UNKNOWN")
        error = item["error_msg"] if item["status"] in TestReportReader.FAILING_TESTS_STATUSES else None
        trname = item.get("workerID", "")
        byte_count = None
        return tstmp, label, concur, rtm, cnn, ltc, rcd, error, trname, byte_count

    def _read(self, last_pass=False):
        for row in self.report_reader.read(last_pass):
            self.read_records += 1
            sample = self.extract_sample(row)
            if sample:
                yield sample


class FuncSamplesReader(FunctionalResultsReader):
    FIELDS_EXTRACTED_TO_ARTIFACTS = ["requestBody", "responseBody", "requestCookiesRaw"]

    def __init__(self, filename, engine, parent_logger):
        self.report_reader = TestReportReader(filename, parent_logger)
        self.engine = engine
        self.read_records = 0

    def _write_sample_data_to_artifacts(self, sample_extras):
        if not sample_extras:
            return
        for file_field in self.FIELDS_EXTRACTED_TO_ARTIFACTS:
            if file_field not in sample_extras:
                continue
            contents = sample_extras.pop(file_field)
            if contents:
                filename = "sample-%s" % file_field
                artifact = self.engine.create_artifact(filename, ".bin")
                with open(artifact, 'wb') as fds:
                    fds.write(contents.encode('utf-8'))
                sample_extras[file_field] = artifact

    def _samples_from_row(self, row):
        result = []
        subsamples = [sample for item in row.get("subsamples", []) for sample in self._samples_from_row(item)]
        if any(subsample.get_type() == 'transaction' for subsample in subsamples):
            result.extend([sub for sub in subsamples if sub.get_type() == 'transaction'])
        else:
            sample = FunctionalSample(test_case=row["test_case"], test_suite=row["test_suite"],
                                      status=row["status"], start_time=row["start_time"], duration=row["duration"],
                                      error_msg=row["error_msg"], error_trace=row["error_trace"],
                                      extras=row.get("extras", {}), subsamples=subsamples, path=row.get("path", []))
            result.append(sample)
        return result

    def read(self, last_pass=False):
        for row in self.report_reader.read(last_pass):
            self.read_records += 1
            samples = self._samples_from_row(row)
            for sample in samples:
                self._write_sample_data_to_artifacts(sample.extras)
                yield sample



#! /usr/bin/env python
"""
Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import codecs
import itertools
import json
import logging
import os
import re
import sys
import traceback
from collections import namedtuple
from copy import deepcopy
from optparse import OptionParser

from cssselect import GenericTranslator

from bzt import TaurusInternalException
from bzt.cli import CLI
from bzt.engine import Configuration, EXEC
from bzt.jmx import JMX
from bzt.utils import get_full_path
from bzt.requests_model import has_variable_pattern

INLINE_JSR223_MAX_LEN = 10

KNOWN_TAGS = ["hashTree", "jmeterTestPlan", "TestPlan", "ResultCollector",
              "HTTPSamplerProxy",
              "ThreadGroup",
              "com.blazemeter.jmeter.threads.concurrency.ConcurrencyThreadGroup",
              "kg.apc.jmeter.threads.SteppingThreadGroup",
              "DNSCacheManager",
              "HeaderManager",
              "CacheManager",
              "CookieManager",
              "ConfigTestElement",
              "DurationAssertion",
              "ConstantTimer",
              "RegexExtractor",
              "HtmlExtractor",
              "BoundaryExtractor",
              "com.atlantbh.jmeter.plugins.jsonutils.jsonpathextractor.JSONPathExtractor",
              "com.atlantbh.jmeter.plugins.jsonutils.jsonpathassertion.JSONPathAssertion",
              "XPathAssertion",
              "XPathExtractor",
              "ResponseAssertion",
              "CSVDataSet",
              "GenericController",
              "ResultCollector",
              "OnceOnlyController",
              "Arguments",
              "IfController",
              "LoopController",
              "WhileController",
              "InterleaveControl",
              "RunTime",
              "ForeachController",
              "TransactionController",
              "JSR223PreProcessor",
              "JSR223PostProcessor",
              "BeanShellPreProcessor",
              "BeanShellPostProcessor",
              "JSONPostProcessor",
              "AuthManager",
              JMX.SET_VAR_ACTION,
              JMX.THR_TIMER,
              ]

LOWER_KNOWN_TAGS = [tag.lower() for tag in KNOWN_TAGS]


class JMXasDict(JMX):
    """
    Model
    """

    def __init__(self, log):
        super(JMXasDict, self).__init__()
        self.log = log.getChild(self.__class__.__name__)
        self.global_objects = []
        self.scenario = {EXEC: None, "scenarios": None}
        self.additional_files = {}  # dict(filename -> file content)

    def load(self, original):
        super(JMXasDict, self).load(original)
        self._clean_jmx_tree(self.tree)
        self._get_global_objects()

    def _get_bool_prop(self, element, prop_name):
        """
        Gets bool prop from element
        :param element:
        :param prop_name:
        :return:
        """
        prop_element = element.find(".//boolProp[@name='" + prop_name + "']")
        if prop_element is not None and prop_element.text:
            if prop_element.text.lower() == 'true':
                return True
            elif prop_element.text.lower() == 'false':
                return False
        else:
            self.log.debug("boolProp %s was not found in %s element!", prop_name, element.tag)
            return None

    def _get_string_prop(self, element, prop_name, default=None):
        """
        Gets string prop from element
        :param element:
        :param prop_name:
        :return:
        """
        prop_element = element.find(".//stringProp[@name='" + prop_name + "']")
        if prop_element is not None and prop_element.text:
            return prop_element.text
        else:
            self.log.debug("stringProp %s was not found in %s element!", prop_name, element.tag)
            return default

    def _get_concurrency(self, element):
        """
        concurrency option in tg execution settings
        :return:
        """
        if element.tag == "ThreadGroup":
            concurrency_tag_name = 'ThreadGroup.num_threads'
        else:
            concurrency_tag_name = 'TargetLevel'
        concurrency = self._get_option_string_with_default(element, concurrency_tag_name, "concurrency", 1)
        self.log.debug('Got %s for concurrency in %s (%s)', concurrency, element.tag, element.get("testname"))
        return concurrency

    def _get_ramp_up(self, element):
        """
        ramp_up option in tg settings
        :param element:
        :return:
        """
        if element.tag == "ThreadGroup":
            ramp_up = self._get_option_string_with_default(element, 'ThreadGroup.ramp_time', "ramp-up", 1)
        else:
            unit = self._get_string_prop(element, "Unit")
            ramp_up = self._get_option_string_with_default(element, 'RampUp', "ramp-up", 1)
            if unit == "M":
                ramp_up["ramp-up"] = ramp_up["ramp-up"] * 60
        self.log.debug('Got %s for rampup in %s (%s)', ramp_up, element.tag, element.get("testname"))
        return ramp_up

    def _get_iterations(self, element):
        """
        iteration option in tg execution settings
        :param element:
        :return:
        """
        controller_element = element.find('.//elementProp')
        iterations = self._get_option_string_with_default(controller_element, 'LoopController.loops', "iterations", 1)
        self.log.debug('Got %s for iterations in %s (%s)', iterations, element.tag, element.get("testname"))
        return iterations

    def _get_duration(self, tg_element):
        """
        get hold_for and ramp_up from tg element.
        :param tg_element:
        :return:
        """
        result = {}
        ramp_up = self._get_ramp_up(tg_element)
        if self._get_bool_prop(tg_element, 'ThreadGroup.scheduler'):
            duration_element = tg_element.find("*[@name='ThreadGroup.duration']")
            if duration_element is not None and duration_element.text and duration_element.text.isdigit():
                duration = int(duration_element.text)
                if ramp_up:
                    result.update({"hold-for": str(duration - ramp_up["ramp-up"]) + "s"})
                else:
                    result.update({"hold-for": str(duration) + "s"})
        if ramp_up:
            ramp_up["ramp-up"] = str(ramp_up["ramp-up"]) + "s"
            result.update(ramp_up)

        if tg_element.tag == "com.blazemeter.jmeter.threads.concurrency.ConcurrencyThreadGroup":
            unit = self._get_string_prop(tg_element, "Unit")
            hold_for = self._get_option_string_with_default(tg_element, 'Hold', "hold-for", "60s")
            if unit == "M":
                hold_for["hold-for"] = hold_for["hold-for"] * 60
            hold_for["hold-for"] = str(hold_for["hold-for"]) + "s"
            result.update(hold_for)

        self.log.debug('Got %s for duration in %s (%s)', result, tg_element.tag, tg_element.get("testname"))
        return result

    def _get_throughput(self, element):
        """
        Gets throughput from variable throughput timer
        :param element:
        :return:
        """
        result = {}
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            property_pattern = "kg.apc.jmeter.timers.VariableThroughputTimer"
            timer_elements = [element for element in hashtree.iterchildren() if element.tag == property_pattern]
            if timer_elements:
                load_profile = timer_elements[0].find(".//collectionProp[@name='load_profile']")
                if load_profile is not None:
                    col_props = load_profile.findall(".//collectionProp")
                    if col_props:
                        col_prop = col_props[-1]
                        st_prop = col_prop.findall(".//stringProp")
                        throughput = st_prop[-2].text
                        if throughput and throughput.isdigit():
                            result["throughput"] = int(throughput)
                            self.log.debug('Got %s for throughput in %s (%s)', throughput, element.tag,
                                           element.get("testname"))
        return result

    def _get_option_string_with_default(self, element, prop_name, opt_name, default):
        """
        :param element:
        :return: dict
        """
        result = {}
        if element is not None:
            prop_value = self._get_string_prop(element, prop_name)
            if prop_value and prop_value.isdigit() and int(prop_value) != default:
                result[opt_name] = int(prop_value)
        return result

    def _get_request_body(self, element, request_config):
        """
        Get body params from sampler
        :param element:
        :return: dict
        """
        raw_body = self._get_bool_prop(element, 'HTTPSampler.postBodyRaw')
        query = 'elementProp[name="HTTPsampler.Arguments"]>collectionProp>elementProp'
        xpath = GenericTranslator().css_to_xpath(query)
        if raw_body:
            http_args_element = element.xpath(xpath)[0]
            body = self._get_string_prop(http_args_element, 'Argument.value')
            if body:
                self.log.debug('Got %s for body in %s (%s)', body, element.tag, element.get("name"))
                return {'body': body}
            else:
                return {}
        else:
            url = request_config.get('url', '')
            method = request_config.get('method', 'get')
            return self._get_params(element, xpath, url=url, method=method)

    def _get_params(self, element, xpath, url='', method='get'):
        request_params = {}
        body_params = {}
        http_args_collection = element.xpath(xpath)
        additional_url = ''
        for param in http_args_collection:
            name = param.get("name")
            val = self._get_string_prop(param, 'Argument.value')
            if val is None:
                val = ''
            incompat = self._get_param_incompat(param, val)
            if incompat:
                if method.lower() == 'get':
                    param_str = name
                    if self._get_bool_prop(param, "HTTPArgument.use_equals"):
                        param_str += '='
                    param_str += val
                    msg = "Parameter '%s'='%s' has unsupported option (%s) and added to url: '%s'"
                    self.log.debug(msg, name, val, incompat, param_str)
                    additional_url += '&' + param_str
                    continue  # avoid adding this parameter to body part
                else:
                    msg = "Parameter '%s'='%s' isn't fully supported for %s request: %s"
                    self.log.warning(msg, name, val, method, incompat)

            body_params[name] = val

        if additional_url:
            if url:
                url += '?'
            url += additional_url[1:]

        request_params['url'] = url

        if body_params:
            self.log.debug('Got %s for parameters in %s (%s)', body_params, element.tag, element.get("name"))
            request_params["body"] = body_params

        return request_params

    def _get_param_incompat(self, param, val):
        """
        check if parameter can be processed by standard way (see jmx.py _add_body_from_script)
         or it must be joined with url string
        """
        if not self._get_bool_prop(param, "HTTPArgument.always_encode"):
            return 'always_encode is off'
        if not val and self._get_bool_prop(param, "HTTPArgument.use_equals"):
            return 'use_equals is on for empty value'

    def _get_upload_files(self, element):
        """
        Extract upload files from element
        :param element:
        :return: dict
        """
        query = 'elementProp[name="HTTPsampler.Files"]>collectionProp'
        xpath = GenericTranslator().css_to_xpath(query)
        colls = element.xpath(xpath)
        if not colls:
            return {}

        upload_files = []
        for elem in colls[0]:
            path = self._get_string_prop(elem, 'File.path')
            param = self._get_string_prop(elem, 'File.paramname')
            mime = self._get_string_prop(elem, 'File.mimetype')
            if path is None or param is None:
                continue
            upload_files.append({'param': param, 'path': path, 'mime-type': mime})

        return {'upload-files': upload_files}

    def _get_headers(self, element):
        """
        Get local request headers
        :return:
        """
        headers = {}
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            headers_elements = [element for element in hashtree.iterchildren() if element.tag == "HeaderManager"]
            for headers_element in headers_elements:
                if headers_element is not None:
                    for header in headers_element.find(".//collectionProp").findall(".//elementProp"):
                        header_name = self._get_string_prop(header, 'Header.name')
                        header_value = self._get_string_prop(header, 'Header.value')
                        if header_name and header_value:
                            headers[header_name] = header_value
        if headers:
            self.log.debug('Got %s for headers in %s (%s)', headers, element.tag, element.get("testname"))
            return {"headers": headers}
        else:
            return headers

    def _get_store_cache(self, element):
        """
        store-cache option
        :param element:
        :return:
        """
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            cache_managers = [element for element in hashtree.iterchildren() if element.tag == "CacheManager"]
            if cache_managers:
                self.log.debug('Got %s for cache_managers in %s (%s)', True, element.tag, element.get("testname"))
                return {"store-cache": True}
        return {}

    def _get_store_cookie(self, element):
        """
        store-cookie option
        :param element:
        :return:
        """
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            cookie_managers = [element for element in hashtree.iterchildren() if element.tag == "CookieManager"]
            if cookie_managers:
                self.log.debug('Got %s for cookie_managers in %s (%s)', True, element.tag, element.get("testname"))
                return {"store-cookie": True}
        return {}

    def _get_dns_mgr(self, element):
        """
        use-dns-cache-mgr option
        :param element:
        :return:
        """
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            dns_managers = [element for element in hashtree.iterchildren() if element.tag == "DNSCacheManager"]
            if dns_managers:
                self.log.debug('Got %s for dns_managers in %s (%s)', True, element.tag, element.get("testname"))
                return {"use-dns-cache-mgr": True}
        return {}

    def __get_constant_timer(self, element):
        """
        think-time option
        :param element:
        :return:
        """
        timer = {}
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            timer_element = [element for element in hashtree.iterchildren() if element.tag == "ConstantTimer"]
            if timer_element:
                timer_delay = self._get_string_prop(timer_element[0], 'ConstantTimer.delay')
                if timer_delay:
                    timer = {"think-time": timer_delay + "ms"}
                    self.log.debug('Got %s for timer in %s (%s)', timer_delay, element.tag, element.get("testname"))
        return timer

    def _get_http_request_defaults(self, element):
        """
        timeout
        default-address
        keepalive
        retrieve-resources
        concurrent-pool-size
        :param element:
        :return:
        """
        request_defaults = {}
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            http_defaults = [element for element in hashtree.iterchildren() if element.tag == "ConfigTestElement"]
            if http_defaults:
                http_defaults = http_defaults[0]
                url_info = self._extract_url_info(http_defaults)
                if url_info:
                    default_address = self._make_url(url_info)
                    if default_address:
                        request_defaults["default-address"] = default_address
                if url_info.timeout:
                    request_defaults["timeout"] = url_info.timeout + "ms"
                if url_info.retrieve_resources is not None:
                    request_defaults["retrieve-resources"] = url_info.retrieve_resources
                    if url_info.retrieve_resources:
                        if url_info.retrieve_concurrency and url_info.retrieve_concurrency.isdigit():
                            request_defaults["concurrent-pool-size"] = int(url_info.retrieve_concurrency)
                if url_info.content_encoding is not None:
                    request_defaults["content-encoding"] = url_info.content_encoding
        self.log.debug('Got %s for request-defaults in %s (%s)', request_defaults, element.tag, element.get("testname"))
        return request_defaults

    @staticmethod
    def _make_url(url_info):
        """
        :type url_info: urllib.ParseResults
        :return: string
        """
        path = "/" if not url_info.path else url_info.path
        port = "" if not url_info.port or url_info.port == "80" else ":" + url_info.port
        protocol = "http" if not url_info.protocol else url_info.protocol
        if url_info.domain:
            return protocol + "://" + url_info.domain + port + path
        return path

    def _extract_url_info(self, element):
        """
        extracts domain, port, etc from element
        :return:
        """
        http_sampler_info = namedtuple("http_sampler_info",
                                       ["domain", "port", "timeout", "protocol", "path", "method", "retrieve_resources",
                                        "retrieve_concurrency", "content_encoding"])
        if element is not None:
            domain = self._get_string_prop(element, 'HTTPSampler.domain')
            port = self._get_string_prop(element, 'HTTPSampler.port')
            timeout = self._get_string_prop(element, 'HTTPSampler.connect_timeout')
            protocol = self._get_string_prop(element, 'HTTPSampler.protocol')
            path = self._get_string_prop(element, 'HTTPSampler.path')
            retrieve_resources = self._get_bool_prop(element, 'HTTPSampler.image_parser')
            retrieve_concurrency = self._get_string_prop(element, 'HTTPSampler.concurrentPool')
            method = self._get_string_prop(element, 'HTTPSampler.method')
            encoding = self._get_string_prop(element, 'HTTPSampler.contentEncoding')
            url_info = http_sampler_info(domain, port, timeout, protocol, path, method, retrieve_resources,
                                         retrieve_concurrency, encoding)
            return url_info
        return None

    def _get_request_base(self, element):
        """
        Base request settings
        :return:
        """
        base_settings = {}
        url_info = self._extract_url_info(element)
        if url_info is not None:
            full_url = self._make_url(url_info)
            base_settings["url"] = full_url
            if url_info.method:
                base_settings["method"] = url_info.method
            if url_info.content_encoding:
                base_settings["content-encoding"] = url_info.content_encoding
        if element.get("testname"):
            base_settings["label"] = element.get("testname")
        self.log.debug('Got %s for base settings in %s (%s)', base_settings, element.tag, element.get("testname"))
        return base_settings

    def _get_authorization(self, element):
        authorization = {}
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            auth_mgrs = [element for element in hashtree.iterchildren() if element.tag == "AuthManager"]

            num_auth_mgrs = len(auth_mgrs)
            if not num_auth_mgrs:
                return {}
            elif num_auth_mgrs > 1:
                self.log.warning('Found %s Authorization Manager elements, only first one will be used')

            auth_mgr = auth_mgrs[0]

            clear_flag = self._get_bool_prop(auth_mgr, "AuthManager.clearEachIteration")
            if clear_flag:
                authorization['clear'] = clear_flag

            auth_list = []

            selector = ".//collectionProp[@name='AuthManager.auth_list']"
            auth_collection = auth_mgr.find(selector)

            if auth_collection is None or not auth_collection.findall(".//elementProp"):
                self.log.warning("Authorizations collection not found in %s, skipping", auth_mgr.tag)
                return {}

            for line in auth_collection.findall(".//elementProp"):
                auth_element = {}

                props = {
                    "url": line.find("stringProp[@name='Authorization.url']"),
                    "name": line.find("stringProp[@name='Authorization.username']"),
                    "password": line.find("stringProp[@name='Authorization.password']"),
                    "domain": line.find("stringProp[@name='Authorization.domain']"),
                    "realm": line.find("stringProp[@name='Authorization.realm']"),
                    "mechanism": line.find("stringProp[@name='Authorization.mechanism']")}

                for key in props:
                    if props[key] is not None:
                        text = props[key].text
                        if text:
                            auth_element[key] = text

                auth_list.append(auth_element)

            authorization["list"] = auth_list
            msg = "Got %s for authorization in %s (%s)"
            self.log.debug(msg, authorization, auth_mgr.tag, auth_mgr.get("testname"))
            return {"authorization": authorization}

        return {}

    def _get_data_sources(self, element, recursive=False):
        """
        data-sources option
        :param element:
        :return: list of dicts
        """
        data_sources = []
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            if recursive:
                elements = hashtree.iterdescendants()
            else:
                elements = hashtree.iterchildren()

            data_sources_elements = [elem for elem in elements if elem.tag == "CSVDataSet"]
            for data_source in data_sources_elements:
                self.log.debug("datasource file: %s", data_source.get("testname"))
                if data_source is not None:
                    data_source_dict = {}
                    f_name_prop = self._get_string_prop(data_source, 'filename')
                    if f_name_prop:
                        data_source_dict["path"] = f_name_prop
                    else:
                        self.log.warning("File name was not set in %s, skipping", data_source.tag)
                        continue
                    delimiter_prop = self._get_string_prop(data_source, 'delimiter')
                    if delimiter_prop:
                        data_source_dict["delimiter"] = delimiter_prop
                    else:
                        self.log.warning("Delimiter was not set in %s, using default: ','", data_source.tag)
                        data_source_dict["delimiter"] = ","
                    quoted_prop = self._get_bool_prop(data_source, 'quotedData')
                    if quoted_prop is not None:
                        data_source_dict["quoted"] = quoted_prop
                    else:
                        self.log.warning("Quoted property was not set in %s, using default: False", data_source.tag)
                        data_source_dict["quoted"] = False

                    loop_prop = self._get_bool_prop(data_source, 'recycle')
                    if loop_prop is None:
                        self.log.warning("Loop property was not set in %s, using default: False", data_source.tag)
                        loop_prop = False

                    stop_prop = self._get_bool_prop(data_source, 'stopThread')
                    if stop_prop is None:
                        self.log.warning("'Stop Thread on EOF' property was not set in %s, using default: False",
                                         data_source.tag)
                        stop_prop = False

                    if loop_prop:
                        data_source_dict["loop"] = True
                    elif stop_prop:
                        data_source_dict["loop"] = False

                    varnames = self._get_string_prop(data_source, 'variableNames')
                    if varnames:
                        data_source_dict["variable-names"] = varnames

                    data_sources.append(data_source_dict)
        if data_sources:
            self.log.debug('Got %s for data_sources in %s (%s)', data_sources, element.tag, element.get("testname"))
            return {"data-sources": data_sources}
        else:
            return {}

    def _get_request_timeout(self, element):
        timeout = {}
        timeout_prop = self._get_string_prop(element, 'HTTPSampler.connect_timeout')
        if timeout_prop:
            self.log.debug('Got %s for request timeout in %s (%s)', timeout_prop, element.tag, element.get("testname"))
            timeout = {"timeout": timeout_prop + "ms"}
        return timeout

    def _get_request_redirect_policy(self, element):
        redirect = {}
        follow_redirects = self._get_bool_prop(element, 'HTTPSampler.follow_redirects')
        auto_redirects = self._get_bool_prop(element, 'HTTPSampler.auto_redirects')
        if not follow_redirects and not auto_redirects:
            redirect["follow-redirects"] = False
        elif follow_redirects or auto_redirects:
            redirect["follow-redirects"] = True
        # NOTE: there's no need to handle 'if follow_redirects is True or auto_redirects is True' case
        # as it's the default behaviour
        return redirect

    def _get_extractors(self, element):
        """
        Gets xpath, jsonpath, regexp and html extractors
        :param element:
        :return:
        """
        extractors = {}
        regexp_extractors = self._get_regexp_extractor(element)
        if regexp_extractors:
            extractors.update({"extract-regexp": regexp_extractors})
        jsonpath_extractors = self._get_json_path_extractors(element)
        if jsonpath_extractors:
            extractors.update({"extract-jsonpath": jsonpath_extractors})
        xpath_extractors = self._get_xpath_extractors(element)
        if xpath_extractors:
            extractors.update({"extract-xpath": xpath_extractors})
        boundary_extractors = self._get_boundary_extractors(element)
        if boundary_extractors:
            extractors.update({"extract-boundary": boundary_extractors})
        html_extractors = self._get_html_extractors(element)
        if html_extractors:
            extractors.update({"extract-css-jquery": html_extractors})
        self.log.debug('Got %s for extractors in %s (%s)', extractors, element.tag, element.get("testname"))
        return extractors

    def _get_regexp_extractor(self, element):
        """
        extract-regexp option
        :param element:
        :return:
        """
        regexp_extractors = {}
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            extractor_elements = [element for element in hashtree.iterchildren() if element.tag == "RegexExtractor"]
            for extractor_element in extractor_elements:
                regexp_extractor = {}
                if extractor_element is not None:
                    refname = self._get_string_prop(extractor_element, 'RegexExtractor.refname')

                    if refname:
                        extractor_props = {}
                        regexp_prop = self._get_string_prop(extractor_element, 'RegexExtractor.regex')
                        if regexp_prop:
                            extractor_props["regexp"] = regexp_prop
                        else:
                            self.log.warning("No regexp expression found in %s, skipping", extractor_element.tag)
                            continue

                        default_prop = self._get_string_prop(extractor_element, 'RegexExtractor.default')

                        if default_prop:
                            extractor_props["default"] = default_prop
                        else:
                            self.log.warning("No default value found in %s", extractor_element.tag)
                            extractor_props["default"] = ""

                        match_no_prop = self._get_string_prop(extractor_element, 'RegexExtractor.match_number')

                        if match_no_prop and match_no_prop.isdigit():
                            extractor_props["match-no"] = int(match_no_prop)
                        else:
                            self.log.warning("No match number found in %s, using default: 0", extractor_element.tag)
                            extractor_props["match-no"] = 0

                        template_prop = self._get_string_prop(extractor_element, 'RegexExtractor.template')

                        if template_prop:
                            extractor_props["template"] = template_prop
                        else:
                            self.log.warning("No template property found in %s, using default: $0$",
                                             extractor_element.tag)
                            extractor_props["template"] = '$0$'

                        regexp_extractor.update({refname: extractor_props})
                    else:
                        self.log.warning("refname property element not found in %s skipping", extractor_element.tag)
                        continue
                regexp_extractors.update(regexp_extractor)

        return regexp_extractors

    def _get_boundary_extractors(self, element):
        """
        extract-boundary option
        :param element:
        :return:
        """
        boundary_extractors = {}
        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            extractor_elements = [element for element in hashtree.iterchildren() if element.tag == "BoundaryExtractor"]
            for extractor_element in extractor_elements:
                if extractor_element is None:
                    continue

                refname = self._get_string_prop(extractor_element, 'BoundaryExtractor.refname')
                if refname is None:
                    self.log.warning("refname property element not found in %s skipping", extractor_element.tag)
                    continue

                extractor_props = {}
                left = self._get_string_prop(extractor_element, 'BoundaryExtractor.lboundary')
                right = self._get_string_prop(extractor_element, 'BoundaryExtractor.rboundary')
                if left and right:
                    extractor_props["left"] = left
                    extractor_props["right"] = right
                else:
                    self.log.warning("Incomplete boundaries in %s, skipping", extractor_element.tag)
                    continue

                default_prop = self._get_string_prop(extractor_element, 'BoundaryExtractor.default')

                if default_prop:
                    extractor_props["default"] = default_prop
                else:
                    self.log.warning("No default value found in %s", extractor_element.tag)
                    extractor_props["default"] = ""

                match_no_prop = self._get_string_prop(extractor_element, 'BoundaryExtractor.match_number')

                if match_no_prop and match_no_prop.isdigit():
                    extractor_props["match-no"] = int(match_no_prop)
                else:
                    self.log.warning("No match number found in %s, using default: 0", extractor_element.tag)
                    extractor_props["match-no"] = 0

                subject = self._get_string_prop(extractor_element, 'BoundaryExtractor.useHeaders')

                subjects = {
                    'false': 'body',
                    'unescaped': 'body-unescaped',
                    'as_document': 'body-as-document',
                    'true': 'response-headers',
                    'request_headers': 'request-headers',
                    'url': 'url',
                    'code': 'code',
                    'message': 'message',
                }
                if subject and subject.lower() in subjects:
                    extractor_props["subject"] = subjects.get(subject.lower())
                else:
                    self.log.warning("No useHeaders property found in %s, using default: body",
                                     extractor_element.tag)
                    extractor_props["subject"] = 'body'

                boundary_extractors.update({refname: extractor_props})

        return boundary_extractors

    def _get_html_extractors(self, element):
        """
        extract-css-jquery option
        :param element:
        :return:
        """
        html_extractors = {}
        hashtree = element.getnext()

        if hashtree is not None and hashtree.tag == "hashTree":
            extractor_elements = [element for element in hashtree.iterchildren() if element.tag == "HtmlExtractor"]

            for extractor_element in extractor_elements:
                if extractor_element is None:
                    continue

                refname = self._get_string_prop(extractor_element, "HtmlExtractor.refname")
                if refname is None:
                    self.log.warning("refname property element not found in %s skipping", extractor_element.tag)
                    continue

                extractor_props = {}

                expression_prop = self._get_string_prop(extractor_element, "HtmlExtractor.expr")
                if expression_prop:
                    extractor_props["expression"] = expression_prop
                else:
                    self.log.warning("No html expression found in %s, skipping element", extractor_element.tag)
                    continue

                attribute_prop = self._get_string_prop(extractor_element, "HtmlExtractor.attribute")
                if attribute_prop:
                    extractor_props["attribute"] = attribute_prop
                else:
                    self.log.warning("No hmtl attribute found in %s, skipping element", extractor_element.tag)
                    continue

                match_no_prop = self._get_string_prop(extractor_element, "HtmlExtractor.match_number")
                if match_no_prop:
                    extractor_props["match-no"] = int(match_no_prop)
                else:
                    self.log.warning("No match number found in %s, using default: 0", extractor_element.tag)
                    extractor_props["match-no"] = 0

                default_prop = self._get_string_prop(extractor_element, "HtmlExtractor.default")
                if default_prop:
                    extractor_props["default"] = default_prop
                else:
                    self.log.warning("No default value found in %s", extractor_element.tag)
                    extractor_props["default"] = ""

                html_extractors.update({refname: extractor_props})

        return html_extractors

    def _get_json_path_extractors(self, element):
        """
        extract-jsonpath option
        :param element:
        :return:
        """

        json_path_extractors = {}

        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            plugins_extractor_pattern = "com.atlantbh.jmeter.plugins.jsonutils.jsonpathextractor.JSONPathExtractor"
            native_extractor_pattern = "JSONPostProcessor"
            extractor_elements = [
                element
                for element in hashtree.iterchildren()
                if element.tag in (plugins_extractor_pattern, native_extractor_pattern)
            ]
            for extractor_element in extractor_elements:
                if extractor_element is None:
                    continue
                if extractor_element.tag == plugins_extractor_pattern:
                    varname = self._get_string_prop(extractor_element, 'VAR')
                    if varname:
                        extractor_props = {}
                        jsonpath_prop = self._get_string_prop(extractor_element, 'JSONPATH')

                        if jsonpath_prop:
                            extractor_props["jsonpath"] = jsonpath_prop
                        else:
                            self.log.warning("No json expression found in %s, skipping element", extractor_element.tag)
                            continue

                        default_prop = self._get_string_prop(extractor_element, 'DEFAULT')

                        if default_prop:
                            extractor_props["default"] = default_prop
                        else:
                            self.log.warning("No default value found in %s", extractor_element.tag)
                            extractor_props["default"] = ""

                        json_path_extractors.update({varname: extractor_props})
                    else:
                        self.log.warning("Not found varname in %s, skipping", extractor_element.tag)
                        continue
                elif extractor_element.tag == native_extractor_pattern:
                    self.log.warning("Found native JSONPath extractor")
                    variables = self._get_string_prop(extractor_element, 'JSONPostProcessor.referenceNames')
                    if not variables:
                        self.log.warning("No vars declared for JSONPath extractor")
                        continue
                    queries = self._get_string_prop(extractor_element, 'JSONPostProcessor.jsonPathExprs')
                    if not variables:
                        self.log.warning("No queries declared for JSONPath extractor")
                        continue
                    def_values = self._get_string_prop(extractor_element, 'JSONPostProcessor.defaultValues')
                    def_values_iter = iter(def_values.split(';') if def_values is not None else [])
                    for var, query in zip(variables.split(';'), queries.split(';')):
                        extractor = {"jsonpath": query}
                        try:
                            extractor["default"] = next(def_values_iter)
                        except StopIteration:
                            extractor["default"] = ""
                        json_path_extractors.update({var: extractor})

        return json_path_extractors

    def _get_xpath_extractors(self, element):
        """
        extract-xpath option
        :param element:
        :return:
        """

        extractors = {}

        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            elements = [element for element in hashtree.iterchildren() if element.tag == "XPathExtractor"]
            for element in elements:
                extractor = {}
                if element is not None:
                    varname = self._get_string_prop(element, 'XPathExtractor.refname')
                    if varname:
                        props = {}
                        xpath = self._get_string_prop(element, 'XPathExtractor.xpathQuery')

                        if xpath:
                            props["xpath"] = xpath
                        else:
                            self.log.warning("No xpath query found in %s, skipping element", element.tag)
                            continue

                        default = self._get_string_prop(element, 'XPathExtractor.default')

                        if default:
                            props["default"] = default
                        else:
                            self.log.warning("No default value found in %s", element.tag)
                            props["default"] = ""

                        validate = self._get_bool_prop(element, 'XPathExtractor.validate')
                        props["validate-xml"] = validate if validate else False
                        whitespace = self._get_bool_prop(element, 'XPathExtractor.whitespace')
                        props["ignore-whitespace"] = whitespace if whitespace else False
                        tolerant = self._get_bool_prop(element, 'XPathExtractor.tolerant')
                        props["use-tolerant-parser"] = tolerant if tolerant else False

                        extractor.update({varname: props})

                    else:
                        self.log.warning("Not found varname in %s, skipping", element.tag)
                        continue

                extractors.update(extractor)

        return extractors

    def _get_assertions(self, element):
        """
        assertions:
        assert, assert-jsonpath, assert-xpath
        :param element:
        :return:
        """
        assertions = {}
        simple_assertions = self._get_response_assertions(element)
        if simple_assertions:
            assertions.update({"assert": simple_assertions})
        jsonpath_assertions = self._get_jsonpath_assertions(element)
        if jsonpath_assertions:
            assertions.update({"assert-jsonpath": jsonpath_assertions})
        xpath_assertions = self._get_xpath_assertions(element)
        if xpath_assertions:
            assertions.update({"assert-xpath": xpath_assertions})
        self.log.debug('Got %s for assertions in %s (%s)', assertions, element.tag, element.get("testname"))
        return assertions

    def _extract_test_field(self, jmx_assertion):
        supported_subjects = {"Assertion.response_data": "body",  # "Text Response"
                              "Assertion.response_headers": "headers",  # "Response Headers"
                              "Assertion.response_code": "http-code"}  # "Response Code"
        unsupported_subjects = {"Assertion.response_data_as_document": "body",  # "Document (text)"
                                "Assertion.sample_label": "body",  # "URL Sampled"
                                "Assertion.response_message": "body"}  # "Response Message"

        test_field_prop = self._get_string_prop(jmx_assertion, 'Assertion.test_field')  # "Response Field to test"

        msg = 'in %s, replaced with Assertion.response_data ("Text response")' % jmx_assertion.tag
        msg = '%s test_field ("Response Field to test") %s ' + msg
        if not test_field_prop:
            self.log.warning(msg, 'Empty', 'provided')
        elif test_field_prop in supported_subjects:
            return supported_subjects[test_field_prop]
        elif test_field_prop in unsupported_subjects:
            self.log.warning(msg, 'Unsupported', test_field_prop)
        else:
            self.log.warning(msg, 'Unknown', test_field_prop)

        return 'body'

    def _extract_test_strings(self, jmx_element):
        selector = ".//collectionProp[@name='Asserion.test_strings']"
        assertion_collection = jmx_element.find(selector)

        if assertion_collection is None:
            self.log.warning("Strings collection not found in %s, skipping", jmx_element.tag)
            return []  # empty list for pylint check in _get_response_assertion()

        test_string_props = assertion_collection.findall(".//stringProp")
        test_strings = []
        for string_prop in test_string_props:
            if string_prop is not None and string_prop.text:
                test_strings.append(string_prop.text)

        if not test_strings:
            self.log.warning("No test strings in %s, skipping", jmx_element.tag)

        return test_strings

    def _extract_test_type(self, jmx_element):
        test_types = {2,  # Contains
                      1,  # Matches
                      8,  # Equals
                      16}  # Substring

        test_type_element = jmx_element.find(".//*[@name='Assertion.test_type']")  # "Pattern Matching Rules"

        if test_type_element is None or not test_type_element.text:
            self.log.warning("No test subject provided in %s, skipping", jmx_element.tag)
            return

        test_type = int(test_type_element.text)
        is_inverted = test_type not in test_types

        if test_type not in test_types:
            test_type -= 4

        if test_type not in test_types:
            self.log.warning("Unknown test type in %s, skipping", jmx_element.tag)
            return

        return is_inverted, test_type

    def _extract_assume_success(self, jmx_element):
        assume_success_element = jmx_element.find(".//*[@name='Assertion.assume_success']")  # "Ignore Status"
        if assume_success_element is None or not assume_success_element.text:
            self.log.warning("No assume_success element provided in %s, skipping", jmx_element.tag)
            return

        return assume_success_element.text == 'true'

    def _extract_scope(self, jmx_element):
        scope = jmx_element.find(".//*[@name='Assertion.scope']")  # "Apply to:"
        if scope is None:
            return "main"
        elif scope.text in ("all", "children", "variable"):
            self.log.warning("Unsupported scope '%s' in %s, changed to 'Main sample only", scope, jmx_element.tag)
            return "main"
        else:
            self.log.warning("Unknown scope '%s' in %s", scope, jmx_element.tag)

    def _get_response_assertion(self, jmx_element):
        """
        Convert JMeter Response Assertion to Taurus config format

        """
        assertion = {}

        assertion_strings = self._extract_test_strings(jmx_element)
        if not assertion_strings:
            return
        assertion["contains"] = assertion_strings

        subject_name = self._extract_test_field(jmx_element)
        if not subject_name:
            return
        assertion["subject"] = subject_name

        test_type = self._extract_test_type(jmx_element)
        if not test_type:
            return
        assertion["not"], test_type = test_type

        assume_success = self._extract_assume_success(jmx_element)
        if assume_success is None:
            return
        assertion["assume-success"] = assume_success

        scope = self._extract_scope(jmx_element)
        if not scope:
            return

        if test_type in (2, 16):  # contains, substring (the synonyms)
            assertion["regexp"] = False
        elif test_type == 1:  # matches
            assertion["regexp"] = True
        elif test_type == 8:  # equal
            self.log.warning("Unsupported rule 'Equals' in %s, converted to 'Matches'", jmx_element.tag)
            assertion["regexp"] = True
            modified_strings = ['^' + re.escape(string) + '$' for string in assertion_strings]
            assertion["contains"] = modified_strings
        else:
            raise TaurusInternalException("Wrong test_type '%s' in %s" % (test_type, jmx_element.tag))
        return assertion

    def _get_response_assertions(self, element):
        """
        :param element:
        :return: list of dicts
        """
        response_assertions = []

        hashtree = element.getnext()

        if hashtree is not None and hashtree.tag == "hashTree":
            response_assertion_elements = [element for element in hashtree.iterchildren() if
                                           element.tag == "ResponseAssertion"]

            for response_assertion_element in response_assertion_elements:
                response_assertion = self._get_response_assertion(response_assertion_element)

                if response_assertion:
                    response_assertions.append(response_assertion)

        return response_assertions

    def _get_jsonpath_assertions(self, element):
        """
        assert-jsonpath option
        :param element:
        :return: list of dicts
        """
        json_path_assertions = []
        hashtree = element.getnext()

        if hashtree is not None and hashtree.tag == "hashTree":
            pattern = "com.atlantbh.jmeter.plugins.jsonutils.jsonpathassertion.JSONPathAssertion"
            json_path_assertion_elements = [element for element in hashtree.iterchildren() if element.tag == pattern]
            for json_path_assertion_element in json_path_assertion_elements:
                json_path_assertion = {}
                json_path_element = self._get_string_prop(json_path_assertion_element, 'JSON_PATH')

                if json_path_element:
                    json_path_assertion["jsonpath"] = json_path_element
                else:
                    self.log.warning("No json path in %s, skipping", json_path_assertion_element.tag)
                    continue

                expected_value_element = self._get_string_prop(json_path_assertion_element, 'EXPECTED_VALUE')

                if expected_value_element:
                    json_path_assertion["expected-value"] = expected_value_element

                validate_element = self._get_bool_prop(json_path_assertion_element, 'JSONVALIDATION')
                json_path_assertion["validate"] = False if validate_element is None else validate_element
                expect_null_element = self._get_bool_prop(json_path_assertion_element, 'EXPECT_NULL')
                json_path_assertion["expect-null"] = False if expect_null_element is None else expect_null_element
                invert_elem = self._get_bool_prop(json_path_assertion_element, 'INVERT')
                json_path_assertion["invert"] = False if invert_elem is None else invert_elem
                regexp = self._get_bool_prop(json_path_assertion_element, 'ISREGEX')
                json_path_assertion["regexp"] = True if regexp is None else regexp

                json_path_assertions.append(json_path_assertion)

        return json_path_assertions

    def _get_xpath_assertions(self, element):
        """
        assert-xpath option
        :param element:
        :return: list of dicts
        """
        assertions = []
        hashtree = element.getnext()

        if hashtree is not None and hashtree.tag == "hashTree":
            assertion_elements = [element for element in hashtree.iterchildren() if element.tag == "XPathAssertion"]
            for assertion_element in assertion_elements:
                assertion = {}
                xpath_element = self._get_string_prop(assertion_element, 'XPath.xpath')

                if xpath_element:
                    assertion["xpath"] = xpath_element
                else:
                    self.log.warning("No xpath in %s, skipping", assertion_element.tag)
                    continue

                validate = self._get_bool_prop(assertion_element, 'XPath.validate')
                assertion["validate-xml"] = validate if validate else False
                whitespace = self._get_bool_prop(assertion_element, 'XPath.whitespace')
                assertion["ignore-whitespace"] = whitespace if whitespace else False
                tolerant = self._get_bool_prop(assertion_element, 'XPath.tolerant')
                assertion["use-tolerant-parser"] = tolerant if tolerant else False
                invert = self._get_bool_prop(assertion_element, 'XPath.negate')
                assertion["invert"] = invert if invert else False

                assertions.append(assertion)

        return assertions

    def _get_variables(self, element):
        """
        :param element:
        :return:
        """
        variables = {}
        hashtree = element.getnext()

        if hashtree is not None and hashtree.tag == "hashTree":
            arguments = [element for element in hashtree.iterchildren() if element.tag == "Arguments"]
            for argument in arguments:
                self.__parse_argument_element(argument, variables)

        return {"variables": variables} if variables else {}

    def __parse_argument_element(self, argument, variables):
        for element in argument.find(".//collectionProp").findall(".//elementProp"):
            var_name = self._get_string_prop(element, 'Argument.name')
            var_value = self._get_string_prop(element, 'Argument.value')
            if var_name and var_value:
                variables[var_name] = var_value

    def _get_jsr223_processors(self, element):
        """
        jsr223 option
        :param element:
        :return:
        """
        extensions = {
            'javascript': '.js',
            'beanshell': '.bsh',
            'bsh': '.bsh',
            'ecmascript': '.js',
            'groovy': '.groovy',
            'java': '.java',
            'jexl': '.jexl',
            'jexl2': '.jexl2',
        }
        jsrs = []

        hashtree = element.getnext()
        if hashtree is not None and hashtree.tag == "hashTree":
            preprocessors = ["JSR223PreProcessor", "BeanShellPreProcessor"]
            postprocessors = ["JSR223PostProcessor", "BeanShellPostProcessor"]
            elements = [element
                        for element in hashtree.iterchildren()
                        if element.tag in preprocessors + postprocessors]
            for element in elements:
                if element is not None:
                    beanshell = element.tag.lower().startswith('beanshell')
                    language = 'beanshell' if beanshell else self._get_string_prop(element, 'scriptLanguage')
                    filename = self._get_string_prop(element, 'filename')
                    params = self._get_string_prop(element, 'parameters')
                    script = self._get_string_prop(element, 'script')
                    cache_key = self._get_string_prop(element, 'cacheKey', default=True)
                    execute = "before" if element.tag in preprocessors else "after"

                    jsr = {"language": language, "parameters": params,
                           "execute": execute, "compile-cache": cache_key}
                    if filename:
                        jsr["script-file"] = filename
                    elif script:
                        if len(script.strip().split('\n')) > INLINE_JSR223_MAX_LEN:
                            ext = extensions.get(language, '.js')
                            filename = self._record_additional_file('script', ext, script)
                            self.additional_files[filename] = script
                            jsr['script-file'] = filename
                        else:
                            jsr['script-text'] = script
                    else:
                        tmpl = "%s element doesn't have neither script nor script-file, skipping"
                        self.log.warning(tmpl, element.tag)
                        continue
                    jsrs.append(jsr)
        if jsrs:
            return {"jsr223": jsrs}
        else:
            return {}

    @staticmethod
    def _get_content_type(headers):
        for header in headers:
            if header.lower() == 'content-type':
                if isinstance(headers[header], str):
                    return headers[header].lower()
                else:
                    return headers[header]

    def process_tg(self, tg_etree_element):
        """
        Get execution and scenario settings for TG
        :param tg_etree_element:
        :return: dict
        """
        self.log.debug("Processing thread group... %s", tg_etree_element.get("testname"))
        tg_scenario_settings = self._get_tg_scenario_settings(tg_etree_element)
        tg_name = tg_etree_element.get("testname")

        if not tg_name:
            tg_name = str(tg_etree_element.__hash__())
        ht_element = tg_etree_element.getnext()

        if ht_element.tag == "hashTree":
            requests = self.__extract_requests(ht_element)

            # convert json-body to string
            scenario_json = False
            if 'headers' in tg_scenario_settings:
                scenario_json = self._get_content_type(tg_scenario_settings['headers']) == 'application/json'

            for request in requests:
                if not ('body' in request and isinstance(request['body'], str)):
                    continue
                request_content_header = 'headers' in request and self._get_content_type(request['headers'])

                if (request_content_header and self._get_content_type(request['headers']) == 'application/json') or \
                        (not request_content_header and scenario_json):

                    # quote jmeter variables
                    body = request['body']
                    pattern = re.compile(r'[^"]\${[a-zA-Z0-9_]+}[^"]')
                    search_res = pattern.search(body)
                    while search_res:
                        body = body[:search_res.start() + 1] + \
                               '"%s"' % search_res.group()[1: -1] + \
                               body[search_res.end() - 1:]
                        search_res = pattern.search(body)
                    if body != request['body']:
                        self.log.debug("Body convertion: '%s' => '%s'", request['body'], body)

                    try:
                        request['body'] = json.loads(body)
                    except (ValueError, TypeError):
                        self.log.warning("Following body cannot be converted into JSON:\n%s", body)

            self.log.debug("Total requests in tg groups: %d", len(requests))
            if not requests:
                self.log.warning("No requests in %s (%s)", tg_etree_element.tag, tg_etree_element.get("testname"))
            tg_scenario_settings["requests"].extend(requests)
        tg_scenario_dict = {tg_name: tg_scenario_settings}
        tg_executions_dict = {"scenario": tg_name}
        tg_executions_dict.update(self._get_tg_execution_settings(tg_etree_element))
        return tg_executions_dict, tg_scenario_dict

    def __extract_requests(self, ht_element):
        requests = []
        children = ht_element.iterchildren()
        while True:
            try:
                elem = next(children)
            except StopIteration:
                break
            if elem.tag == 'IfController':
                hash_tree = next(children)
                if_block = self.__extract_if_controller(elem, hash_tree)
                requests.append(if_block)
            elif elem.tag == 'OnceOnlyController':
                hash_tree = next(children)
                once_block = self.__extract_once_controller(hash_tree)
                requests.append(once_block)
            elif elem.tag == JMX.SET_VAR_ACTION:
                set_variables_block = self.__extract_set_variables(elem)
                requests.append(set_variables_block)
            elif elem.tag == 'LoopController':
                hash_tree = next(children)
                loop_block = self.__extract_loop_controller(elem, hash_tree)
                requests.append(loop_block)
            elif elem.tag == 'WhileController':
                hash_tree = next(children)
                while_block = self.__extract_while_controller(elem, hash_tree)
                requests.append(while_block)
            elif elem.tag == 'ForeachController':
                hash_tree = next(children)
                block = self.__extract_foreach_controller(elem, hash_tree)
                requests.append(block)
            elif elem.tag == 'TransactionController':
                hash_tree = next(children)
                block = self.__extract_trans_controller(elem, hash_tree)
                requests.append(block)
            elif elem.tag == 'HTTPSamplerProxy':
                request = self._get_request_settings(elem)
                requests.append(request)
            elif elem.tag == "hashTree":    # structure isn't recognized, just extract requests from hashTree
                subrequests = self.__extract_requests(elem)
                requests.extend(subrequests)
        return requests

    def __extract_if_controller(self, controller, ht_element):
        condition = self._get_string_prop(controller, "IfController.condition")
        requests = self.__extract_requests(ht_element)
        return {'if': condition, 'then': requests}

    def __extract_set_variables(self, action):
        variables = {}
        self.__parse_argument_element(action, variables)
        return {"set-variables": variables} if variables else {}

    def __extract_once_controller(self, ht_element):
        requests = self.__extract_requests(ht_element)
        return {'once': requests}

    def __extract_loop_controller(self, controller, ht_element):
        # NOTE: we can't rely on LoopController.continue_forever , as it's for some reason always `true` on 4.0
        # NOTE: LoopController.loops may be either stringProp or intProp, depending on version

        strprop = controller.find(".//stringProp[@name='LoopController.loops']")
        if strprop is not None and strprop.text:
            iterations = strprop.text
        else:
            intprop = controller.find(".//intProp[@name='LoopController.loops']")
            if intprop is not None and intprop.text:
                iterations = intprop.text
            else:
                self.log.warning("LoopController.loops has non-numeric value, resetting to 1")
                iterations = 1
        if not has_variable_pattern(iterations):
            try:
                iterations = int(iterations)
            except ValueError:
                msg = "Wrong iterations value in %s Loop Controller: %s, replace with default (1)"
                self.log.warning(msg, controller.get("testname"), iterations)
                iterations = 1

        loops = 'forever' if iterations == -1 else iterations
        requests = self.__extract_requests(ht_element)
        return {'loop': loops, 'do': requests}

    def __extract_while_controller(self, controller, ht_element):
        condition = self._get_string_prop(controller, "WhileController.condition")
        requests = self.__extract_requests(ht_element)
        return {'while': condition, 'do': requests}

    def __extract_foreach_controller(self, controller, ht_element):
        input_var = self._get_string_prop(controller, "ForeachController.inputVal", default="")
        loop_var = self._get_string_prop(controller, "ForeachController.returnVal", default="")
        requests = self.__extract_requests(ht_element)
        iteration_str = '%s in %s' % (loop_var, input_var)
        return {'foreach': iteration_str, 'do': requests}

    def __extract_trans_controller(self, controller, ht_element):
        name = controller.get('testname')
        requests = self.__extract_requests(ht_element)
        include_timers = self._get_bool_prop(controller, "TransactionController.includeTimers")
        if include_timers is None:
            include_timers = True  # weird, but true
        return {'transaction': name, 'do': requests, 'include-timers': include_timers}

    def _get_request_settings(self, request_element):
        """
        Gets all possible request settings
        :param request_element:
        :return: dict
        """
        request_config = {}
        request_config.update(self._get_request_base(request_element))
        request_config.update(self._get_request_body(request_element, request_config))
        request_config.update(self._get_upload_files(request_element))
        request_config.update(self._get_headers(request_element))
        request_config.update(self.__get_constant_timer(request_element))
        request_config.update(self._get_request_timeout(request_element))
        request_config.update(self._get_request_redirect_policy(request_element))
        request_config.update(self._get_extractors(request_element))
        request_config.update(self._get_assertions(request_element))
        request_config.update(self._get_jsr223_processors(request_element))
        return request_config

    def _get_tg_scenario_settings(self, tg_etree_element):
        """
        Gets all possible tg settings and applies global overrides
        :param tg_etree_element:
        :return:
        """
        default_tg_settings = {"store-cookie": False, "store-cache": False, "use-dns-cache-mgr": False}
        global_tg_settings = self._get_global_tg_scenario()
        tg_settings = {"requests": []}
        tg_settings.update(default_tg_settings)
        tg_settings.update(self._get_authorization(tg_etree_element))
        tg_settings.update(self._get_data_sources(tg_etree_element, recursive=True))
        tg_settings.update(self._get_headers(tg_etree_element))
        tg_settings.update(self._get_store_cache(tg_etree_element))
        tg_settings.update(self._get_store_cookie(tg_etree_element))
        tg_settings.update(self._get_dns_mgr(tg_etree_element))
        tg_settings.update(self._get_extractors(tg_etree_element))
        tg_settings.update(self._get_assertions(tg_etree_element))
        tg_settings.update(self._get_variables(tg_etree_element))
        # apply global test plan settings:
        self._apply_global_tg_settings(global_tg_settings, tg_settings)
        # those settings will override global:
        tg_settings.update(self.__get_constant_timer(tg_etree_element))
        tg_settings.update(self._get_http_request_defaults(tg_etree_element))
        return tg_settings

    def _get_tg_execution_settings(self, tg_etree_element):
        """
        Gets execution settings
        :param tg_etree_element:
        :return: dict
        """
        global_execution_settings = self._get_global_tg_execution()
        execution_settings = {
            'concurrency': 1,
            'iterations': 1,
            'ramp-up': '60s',
        }
        self._apply_global_tg_settings(global_execution_settings, execution_settings)

        if tg_etree_element.tag == "ThreadGroup":
            execution_settings.update(self._get_concurrency(tg_etree_element))
            execution_settings.update(self._get_iterations(tg_etree_element))
            execution_settings.update(self._get_duration(tg_etree_element))
        else:
            execution_settings.update(self._get_concurrency(tg_etree_element))
            execution_settings.update(self._get_duration(tg_etree_element))
            execution_settings.pop('iterations')

        execution_settings.update(self._get_throughput(tg_etree_element))

        return execution_settings

    def _get_global_objects(self):
        """
        list of global objects in test plan
        :return:
        """
        self.global_objects = []
        try:
            ht_object = self.tree.find(".//hashTree").find(".//TestPlan").getnext()
        except BaseException:
            raise TaurusInternalException("Bad jmx format")
        for obj in ht_object.iterchildren():
            if obj.tag != 'hashTree' and obj.tag != 'ThreadGroup':
                self.global_objects.append(obj)

    def _get_global_tg_scenario(self):
        """
        :return: dict
        """
        default_tg_settings = {}
        testplan_element = self.tree.find(".//TestPlan")
        default_tg_settings.update(self._get_headers(testplan_element))
        default_tg_settings.update(self._get_data_sources(testplan_element))
        default_tg_settings.update(self._get_store_cache(testplan_element))
        default_tg_settings.update(self._get_store_cookie(testplan_element))
        default_tg_settings.update(self._get_dns_mgr(testplan_element))
        default_tg_settings.update(self.__get_constant_timer(testplan_element))
        default_tg_settings.update(self._get_http_request_defaults(testplan_element))
        default_tg_settings.update(self._get_extractors(testplan_element))
        default_tg_settings.update(self._get_assertions(testplan_element))
        default_tg_settings.update(self._get_variables(testplan_element))
        return default_tg_settings

    def _get_global_tg_execution(self):
        """
        :return: dict
        """
        default_tg_settings = {}
        testplan_element = self.tree.find(".//TestPlan")
        default_tg_settings.update(self._get_throughput(testplan_element))
        return default_tg_settings

    @staticmethod
    def _apply_global_tg_settings(defaults, tg_dict, override=True):
        for default_key, default_value in defaults.items():
            if isinstance(default_value, list):
                if default_key in tg_dict.keys():
                    tg_dict[default_key].extend(deepcopy(default_value))
                else:
                    tg_dict[default_key] = deepcopy(default_value)
            elif isinstance(default_value, dict):
                if default_key in tg_dict.keys():
                    tg_dict[default_key].update(default_value)
                else:
                    tg_dict[default_key] = defaults[default_key]
            else:
                if default_key not in tg_dict.keys():
                    tg_dict[default_key] = defaults[default_key]
                else:
                    if override:
                        tg_dict[default_key] = defaults[default_key]

    def _clean_jmx_tree(self, element):
        """
        Removes all unknown and disabled elements
        :return:
        """
        for subelement in element.findall('./'):
            tag = subelement.tag.lower()
            if tag == 'hashtree':
                self._clean_jmx_tree(subelement)    # look inside
            else:
                unknown = tag not in LOWER_KNOWN_TAGS
                disabled = subelement.get("enabled") == "false"
                if disabled:
                    self.log.info("Removing disabled element: %s (%s)", subelement.tag, subelement.get("testname"))
                elif unknown:
                    self.log.warning("Removing unknown element: %s (%s)", subelement.tag, subelement.get("testname"))
                else:
                    continue

                sibling = subelement.getnext()
                subelement.getparent().remove(subelement)
                next_is_hashtree = sibling is not None and sibling.tag.lower() == "hashtree"

                # remove correspond hashtree
                if next_is_hashtree and (not tag.endswith("controller") or disabled):
                    sibling.getparent().remove(sibling)

                # start from the beginning
                self._clean_jmx_tree(element)
                break

    def _record_additional_file(self, base_filename, extension, content):
        filename = base_filename + extension
        if filename not in self.additional_files:
            self.additional_files[filename] = content
            return filename

        for index in itertools.count(start=1):
            suffix = '-%d' % index
            filename = base_filename + suffix + extension
            if filename not in self.additional_files:
                self.additional_files[filename] = content
                return filename


class Converter(object):
    """
    View
    """

    def __init__(self, log):
        self.log = log.getChild(self.__class__.__name__)
        self.dialect = JMXasDict(self.log)

    def convert(self, file_to_convert, dump_modified_jmx_to=None):
        """
        Get all thread groups from jmx, convert to dict.
        :type dump_modified_jmx_to: str
        :type file_to_convert: str
        :rtype: dict
        """
        self.dialect.load(file_to_convert)
        base_script = {"scenarios": {}, EXEC: []}
        self.log.debug("Processing thread groups...")
        concurrency_tg = ".//com.blazemeter.jmeter.threads.concurrency.ConcurrencyThreadGroup"
        tg_etree_elements = self.dialect.tree.findall(".//ThreadGroup")
        tg_etree_elements.extend(self.dialect.tree.findall(concurrency_tg))
        if not tg_etree_elements:
            raise TaurusInternalException("No thread groups found!")

        if tg_etree_elements:
            self.log.debug("Total thread groups: %d", len(tg_etree_elements))

            for tg_etree_element in tg_etree_elements:
                td_executions_dict, tg_scenario_dict = self.dialect.process_tg(tg_etree_element)
                base_script["scenarios"].update(tg_scenario_dict)
                base_script[EXEC].append(td_executions_dict)
                self.log.debug("Done processing thread group %s", tg_etree_element.get("testname"))

            if dump_modified_jmx_to:
                self._dump_modified_jmx(dump_modified_jmx_to)
            return base_script
        else:
            raise TaurusInternalException("No thread groups was found in JMX file!")

    def _dump_modified_jmx(self, dump_jmx):
        """
        Dump modified JMX from dialect object
        """
        if dump_jmx:
            if os.path.exists(dump_jmx):
                self.log.warning("%s already exists and will be overwritten", dump_jmx)
                self.dialect.save(dump_jmx)
                self.log.info("Cleaned JMX saved in %s", dump_jmx)


class JMX2YAML(object):
    """
    Controller
    """

    def __init__(self, options, file_name):
        self.log = logging.getLogger(self.__class__.__name__)
        self.options = options
        self.setup_logging()
        self.converter = None
        self.src_file = file_name
        self.dst_file = ''

    def setup_logging(self):
        CLI.setup_logging(self.options)
        if self.options.quiet:
            logging.disable(logging.WARNING)

    def process(self):
        """
        Process file
        :return:
        """
        output_format = Configuration.JSON if self.options.json else Configuration.YAML

        self.log.info('Loading jmx file %s', self.src_file)
        self.src_file = os.path.abspath(os.path.expanduser(self.src_file))
        if not os.path.exists(self.src_file):
            raise TaurusInternalException("File does not exist: %s" % self.src_file)
        self.converter = Converter(self.log)
        try:
            jmx_as_dict = self.converter.convert(self.src_file, self.options.dump_jmx)
        except BaseException:
            self.log.error("Error while processing jmx file: %s", self.src_file)
            raise

        exporter = Configuration.from_dict(jmx_as_dict)

        if self.options.file_name:
            self.dst_file = self.options.file_name
        else:
            self.dst_file = self.src_file + "." + output_format.lower()

        with open(self.dst_file, 'wb') as fds:
            exporter.write(fds, output_format)

        additional_files_dir = get_full_path(self.dst_file, step_up=1)
        for filename in self.converter.dialect.additional_files:
            path = os.path.join(additional_files_dir, filename)
            self.log.info("Writing additional file: %s", path)
            content = self.converter.dialect.additional_files[filename]
            with codecs.open(path, 'w', encoding='utf-8') as f:
                f.write(content)

        self.log.info("Done processing, result saved in %s", self.dst_file)


def main():
    usage = "Usage: jmx2yaml [input jmx file] [options]"
    parser = OptionParser(usage=usage, prog="jmx2yaml")
    parser.add_option('-v', '--verbose', action='store_true', default=False,
                      help="Prints all logging messages to console")
    parser.add_option('-o', '--out', dest="file_name",
                      help="Set output .yml file name, by default input file name + .yml is used")
    parser.add_option('-d', '--dump-jmx', action='store', dest='dump_jmx', default=False,
                      help="Dumps processed JMX to file")
    parser.add_option('-q', '--quiet', action='store_true', default=False, dest='quiet',
                      help="Do not display any log messages")
    parser.add_option('-j', '--json', action='store_true', default=False, dest='json',
                      help="Use JSON format")
    parser.add_option('-l', '--log', action='store', default=False, help="Log file location")
    parsed_options, args = parser.parse_args()
    if len(args) > 0:
        tool = JMX2YAML(parsed_options, args[0])
        code = 0
        try:
            tool.process()
        except BaseException as exc:
            logging.error("Exception: %s", exc)
            logging.debug("Exception: %s", traceback.format_exc())
            code = 1
        sys.exit(code)
    else:
        sys.stdout.write(usage + "\n")


if __name__ == "__main__":
    main()



# Utility functions and classes for Taurus Selenium tests
import time

from apiritif import get_transaction_handlers, set_transaction_handlers, get_from_thread_store, get_iteration, external_handler

from selenium.common.exceptions import NoSuchWindowException, NoSuchFrameException, NoSuchElementException, \
    TimeoutException, UnexpectedAlertPresentException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as econd
from selenium.webdriver.remote.webelement import WebElement

from apiritif.action_plugins import BaseActionHandler
from bzt import TaurusException
from bzt.resources.shadow_element import ShadowElement

BYS = {
    'xpath': By.XPATH,
    'css': By.CSS_SELECTOR,
    'name': By.NAME,
    'id': By.ID,
    'linktext': By.LINK_TEXT
}


def find_element_by_shadow(shadow_loc):
    """
    Enables finding element using the Shadow Locator
    :param shadow_loc: shadow locator in the string form - css locators divided by commas, e.g. 'c-comp1, c-basic, .std_btn'
    :return: the found element otherwise NoSuchElementException is raised
    """
    el = None
    css_path = [x.strip() for x in shadow_loc.split(',')]
    for p in css_path:
        if not el:
            el = _get_driver().find_element_by_css_selector(p)
        else:
            curr_el = _find_element_in_shadow(el, p, False)
            if curr_el is None:
                # try to search in the shadowRoot of the parent element
                parent_el = el.find_element_by_xpath('..')
                curr_el = _find_element_in_shadow(parent_el, p, True)
            el = curr_el
    return ShadowElement(el, _get_driver())


def _find_element_in_shadow(el, css_selector, raise_exception):
    shadow_root = el.get_property("shadowRoot")
    element = None
    if shadow_root:
        element = _find_by_css_selector(shadow_root, css_selector, False)
    if element is None:
        # sometimes the element is not located under the shadowRoot so try to look for it the usual way
        element = _find_by_css_selector(el, css_selector, raise_exception)
    return element


def _find_by_css_selector(root, css_selector, raise_exception):
    element = None
    try:
        # Chrome 96+ returns the element as a dict that includes the id
        if isinstance(root, dict):
            key = list(root.keys())[0]
            shadow_root_el = WebElement(_get_driver(), root[key])
            element = shadow_root_el.find_element_by_css_selector(css_selector)
        else:
            element = root.find_element_by_css_selector(css_selector)
    except NoSuchElementException as nse:
        if raise_exception:
            if root.tag_name == "slot":
                raise TaurusException("Shadow DOM Slots are currently not supported in Taurus execution.")
            raise nse
        pass
    return element


def get_locator(locators, parent_el=None, ignore_implicit_wait=False, raise_exception=False):
    """
    :param locators: List of Dictionaries holding the locators, e.g. [{'id': 'elem_id'},
    {css: 'my_cls'}]
    :param parent_el: reference to the parent element (WebElement instance), optional - if provided the find_elements
    method is called on it instead of global context
    :param ignore_implicit_wait: set it to True to set the implicit wait immediately to 0
    :param raise_exception: set it to True to get the NoSuchElementException in case no elements are matching any
    of the passed locators, if set to False then the first locator is returned in that case
    :return: first valid locator from the passed List, if no locator is valid then returns the
    first one
    """
    driver = _get_driver()
    timeout = _get_timeout()
    first_locator = None
    if ignore_implicit_wait:
        driver.implicitly_wait(0)
    if len(locators) == 1 and locators[0].get("shadow"):
        return locators
    for locator in locators:
        locator_type = list(locator.keys())[0]
        locator_value = locator[locator_type]
        if not first_locator:
            first_locator = (BYS[locator_type.lower()], locator_value)
        else:
            # set implicit wait to 0 get the result instantly for the other locators
            driver.implicitly_wait(0)
        if parent_el:
            elements = parent_el.find_elements(BYS[locator_type.lower()], locator_value)
        else:
            elements = driver.find_elements(BYS[locator_type.lower()], locator_value)
        if len(elements) > 0:
            locator = (BYS[locator_type.lower()], locator_value)
            break
    else:
        if raise_exception:
            driver.implicitly_wait(timeout)
            msg = "Element not found: (%s, %s)" % first_locator
            raise NoSuchElementException(msg)
        else:
            locator = first_locator

    # restore the implicit wait value
    driver.implicitly_wait(timeout)
    return locator


def get_elements(locators):
    """
    :param locators: List of Dictionaries holding the locators, e.g. [{'id': 'elem_id'},
    {css: 'my_cls'}]
    :return: all elements that match the first valid locator out of the passed locators
    """
    elements = []
    first_locator = True
    driver = _get_driver()
    for locator in locators:
        locator_type = list(locator.keys())[0]
        locator_value = locator[locator_type]
        if not first_locator:
            driver.implicitly_wait(0)
        elements = driver.find_elements(BYS[locator_type.lower()], locator_value)
        first_locator = False
        if len(elements) > 0:
            break

    # restore the implicit wait value
    driver.implicitly_wait(_get_timeout())
    return elements


def _get_driver():
    return get_from_thread_store("driver")


def _get_timeout():
    timeout = get_from_thread_store("timeout")
    if not (timeout or timeout == 0):  # timeout in (None, []), default requires
        timeout = 30

    return timeout


def add_flow_markers():
    handlers = get_transaction_handlers()
    handlers["enter"].append(_send_start_flow_marker)
    handlers["exit"].append(_send_exit_flow_marker)
    set_transaction_handlers(handlers)


def _send_marker(stage, params):
    _get_driver().execute_script("/* FLOW_MARKER test-case-%s */" % stage, params)


def _send_start_flow_marker(*args, **kwargs):  # for apiritif. remove when compatibiltiy code in
    stage = "start"  # apiritif removed (http.py) and apiritif released ( > 0.9.2)

    test_case, test_suite, scenario_name, data_sources, action_handlers, driver = get_from_thread_store(
        ['test_case', 'test_suite', 'scenario_name', 'data_sources', 'action_handlers', 'driver']
    )
    params = {
        "testCaseName": test_case,
        "testSuiteName": scenario_name or test_suite}

    if data_sources:
        params["testDataIterationId"] = get_iteration()

    for handler in action_handlers:
        handler.handle(driver.session_id, BaseActionHandler.TEST_CASE_START, params)

    _send_marker(stage, params)


def _send_exit_flow_marker(*args, **kwargs):  # for apiritif. remove when compatibiltiy code in
    stage = "stop"  # apiritif removed (http.py) and apiritif released ( > 0.9.2)
    labels = "status", "message"
    action_handlers, driver = get_from_thread_store(['action_handlers', 'driver'])
    values = get_from_thread_store(labels)
    params = dict(zip(labels, values))

    for handler in action_handlers:
        handler.handle(driver.session_id, BaseActionHandler.TEST_CASE_STOP, params)
    _send_marker(stage, params)


def dialogs_replace():
    """
    Replaces the standard JavaScript methods, i.e. 'window.confirm', 'window.alert' and 'window.prompt' with
    own implementation that stores the messages from the dialogs and also is capable of returning user defined
    values
    """

    _get_driver().execute_script("""
          if (window.__webdriverAlerts) { return; }
          window.__webdriverAlerts = [];
          window.__webdriverOriginalAlert = window.alert;
          window.__webdriverNextAlert = null;
          window.alert = function(msg) {
            if (window.__webdriverNextAlert === null) {
                window.__webdriverOriginalAlert(msg);
            }
            window.__webdriverNextAlert = null; 
            window.__webdriverAlerts.push(msg); 
          };
          window.__webdriverConfirms = [];
          window.__webdriverNextConfirm = null;
          window.__webdriverPrevConfirm = window.confirm;
          window.confirm = function(msg) {
            window.__webdriverConfirms.push(msg);
            var res = window.__webdriverNextConfirm;
            if (res === null) {
                return window.__webdriverPrevConfirm(msg);
            }
            window.__webdriverNextConfirm = null;
            return res;
          };
          window.__webdriverPrompts = [];
          window.__webdriverNextPrompts = true;
          window.prompt = function(msg, def) {
            window.__webdriverPrompts.push(msg || def);
            var res = window.__webdriverNextPrompt;
            window.__webdriverNextPrompt = true;
            return res;
          };
        """)


def dialogs_get_next_confirm():
    """
    :return: the message from the last invocation of 'window.confirm'
    """
    return _get_driver().execute_script("""
                 if (!window.__webdriverConfirms) { return null; }
                 return window.__webdriverConfirms.shift();
               """)


def dialogs_get_next_alert():
    """
    :return: the alert message from the last invocation of 'window.alert'
    """
    return _get_driver().execute_script("""
                if (!window.__webdriverAlerts) { return null } 
                var t = window.__webdriverAlerts.shift(); 
                if (t) { t = t.toString().replace(/\\n/g, ' '); }
                return t;
              """)


def dialogs_get_next_prompt():
    """
    :return: the message from the last invocation of 'window.prompt'
    """
    return _get_driver().execute_script("""
                if (!window.__webdriverPrompts) { return null; }
                return window.__webdriverPrompts.shift();
              """)


def dialogs_answer_on_next_alert(value):
    """
    Simulates click on OK button in the next alert
    """
    dialogs_replace()
    if str(value).lower() == '#ok':
        _get_driver().execute_script("window.__webdriverNextAlert = true")


def dialogs_answer_on_next_prompt(value):
    """
    :param value: The value to be used to answer the next 'window.prompt', if '#cancel' is provided then
    click on cancel button is simulated by returning null
    """
    dialogs_replace()
    if str(value).lower() == '#cancel':
        _get_driver().execute_script("window.__webdriverNextPrompt = null")
    else:
        _get_driver().execute_script("window.__webdriverNextPrompt = '%s';" % value)


def dialogs_answer_on_next_confirm(value):
    """
    :param value: either '#ok' to click on OK button or '#cancel' to simulate click on Cancel button in the
    next 'window.confirm' method
    """
    dialogs_replace()
    if str(value).lower() == '#ok':
        confirm = 'true'
    else:
        confirm = 'false'
    _get_driver().execute_script("window.__webdriverNextConfirm = %s;" % confirm)


def wait_for(condition, locators, wait_timeout=10):
    if condition.lower() in ["present", "visible", "clickable"]:
        _wait_for_positive(condition.lower(), locators, wait_timeout)
    elif condition.lower() in ["notpresent", "notvisible", "notclickable"]:
        _wait_for_negative(condition.lower(), locators, wait_timeout)


def _wait_for_positive(condition, locators, wait_timeout):
    start_time = time.time()
    while True:
        locator = None
        try:
            locator = get_locator(locators, ignore_implicit_wait=True, raise_exception=True)
        except NoSuchElementException:
            pass
        if locator:
            element = None
            try:
                element = WebDriverWait(_get_driver(), wait_timeout).until(_get_until_cond(condition, locator))
            except TimeoutException:
                pass
            if element:
                return

        elapsed_time = time.time() - start_time
        if elapsed_time > wait_timeout:
            raise NoSuchElementException("Timeout occurred while waiting for '%s' condition" % condition)


def _wait_for_negative(condition, locators, wait_timeout):
    present_locs = []
    for locator in locators:
        try:
            present_locs.append(get_locator([locator], ignore_implicit_wait=True, raise_exception=True))
        except NoSuchElementException:
            pass
    if not present_locs:
        return
    start_time = time.time()
    for locator in present_locs:
        elapsed_time = time.time() - start_time
        timeout = wait_timeout - elapsed_time
        WebDriverWait(_get_driver(), timeout).until_not(
            _get_until_cond(condition, locator),
            message="Timeout occurred while waiting for element (%s=%s) to become '%s'" %
                    (locator[0], locator[1], condition))


def _get_until_cond(condition, locator):
    loc_tuple = (locator[0], locator[1])
    if "clickable" in condition:
        return econd.element_to_be_clickable(loc_tuple)
    if "present" in condition:
        return econd.presence_of_element_located(loc_tuple)
    if "visible" in condition:
        return econd.visibility_of_element_located(loc_tuple)


def get_loop_range(start, end, step):
    """
    :return: the range over which the loop will operate
    """
    start = int(start)
    end = int(end)
    step = int(step)
    end = end + 1 if step > 0 else end - 1
    return range(start, end, step)


def switch_frame(frame_name=None):
    driver = _get_driver()
    try:
        if not frame_name or frame_name == "relative=top":
            driver.switch_to_default_content()
        elif isinstance(frame_name, str) and frame_name.startswith(
                "index="):  # Switch using index frame using relative position
            driver.switch_to.frame(int(frame_name.split("=")[1]))
        elif frame_name == "relative=parent":  # Switch to parent frame of the current frame
            driver.switch_to.parent_frame()
        else:  # Use the selenium alternative
            driver.switch_to.frame(frame_name)
    except NoSuchFrameException:
        raise NoSuchFrameException("Invalid Frame ID: %s" % frame_name)


def switch_window(window_name=None):
    driver = _get_driver()
    try:
        if window_name is None:  # Switch to last window created
            driver.switch_to.window(driver.window_handles[-1])
        else:
            if isinstance(window_name, int) or window_name.isdigit():  # Switch to window handler index
                _switch_by_idx(int(window_name))
            else:
                if window_name.startswith("win_ser_"):  # Switch using window sequential mode
                    _switch_by_win_ser(window_name)
                else:  # Switch using window name
                    driver.switch_to.window(window_name)
    except NoSuchWindowException:
        raise NoSuchWindowException("Invalid Window ID: %s" % window_name)


def open_window(url):
    """
    Opens the given url in a new window and also switches to it automatically
    """
    driver = _get_driver()
    driver.execute_script("window.open('%s');" % url)
    driver.switch_to.window(driver.window_handles[-1])


def _switch_by_idx(win_index):
    driver = _get_driver()
    wnd_handlers = driver.window_handles
    if 0 <= win_index < len(wnd_handlers):
        driver.switch_to.window(wnd_handlers[win_index])
    else:
        raise NoSuchWindowException("Invalid Window ID: %s" % str(win_index))


def _switch_by_win_ser(window_name):
    driver = _get_driver()
    if window_name == "win_ser_local":
        wnd_handlers = driver.window_handles
        if len(wnd_handlers) > 0:
            driver.switch_to.window(wnd_handlers[0])
        else:
            raise NoSuchWindowException("Invalid Window ID: %s" % window_name)
    elif window_name.split("win_ser_")[1].isdigit():  # e.g. win_ser_1
        _switch_by_idx(int(window_name.split("win_ser_")[1]))
    else:
        windows = get_from_thread_store("windows")
        if window_name not in windows:
            windows[window_name] = driver.current_window_handle
        driver.switch_to.window(windows[window_name])


def close_window(window_name=None):
    if window_name:
        switch_window(window_name)
    _get_driver().close()


def waiter():
    """
    Allows waiting for page to finish loading before performing other actions on non completely loaded page
    """
    try:
        WebDriverWait(_get_driver(), _get_timeout()) \
            .until(lambda driver: driver.execute_script('return document.readyState') == 'complete',
                   message="Timeout occurred while waiting for page to finish loading.")
    except UnexpectedAlertPresentException:
        pass


def action_start(action):
    driver = _get_driver()
    external_handler(driver.session_id if driver else None, BaseActionHandler.YAML_ACTION_START, action)


def action_end(action):
    driver = _get_driver()
    external_handler(driver.session_id if driver else None, BaseActionHandler.YAML_ACTION_END, action)



"""
Module holds all stuff regarding usage of Tsung

Copyright 2016 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import copy
import logging
import os
import re
import traceback
from urllib import parse
from shutil import which

from bzt import TaurusConfigError, ToolError, TaurusInternalException
from bzt.engine import ScenarioExecutor
from bzt.modules.aggregator import ConsolidatingAggregator, ResultsReader
from bzt.modules.console import ExecutorWidget
from bzt.requests_model import HTTPRequest
from bzt.utils import CALL_PROBLEMS, shutdown_process, RequiredTool, dehumanize_time, FileReader, etree, iteritems


class TsungExecutor(ScenarioExecutor):
    """
    Tsung executor module
    """

    def __init__(self):
        super(TsungExecutor, self).__init__()
        self.process = None
        self.__stats_file = None
        self.tsung_config = None
        self.tool = None
        self.tsung_controller_id = None
        self.tsung_artifacts_basedir = None

    def prepare(self):
        super(TsungExecutor, self).prepare()
        scenario = self.get_scenario()
        self.install_required_tools()

        script = self.get_script_path()
        if script:
            if not os.path.exists(script):
                raise TaurusConfigError("Tsung: script '%s' doesn't exist" % script)
            self.tsung_config = self.__modify_user_tsung_config(script)
        elif scenario.get("requests"):
            self.tsung_config = self._generate_tsung_config()
        else:
            raise TaurusConfigError("Tsung: you must specify either a script or a list of requests")

        self.tsung_controller_id = "tsung_taurus_%s" % id(self)
        self.tsung_artifacts_basedir = os.path.join(self.engine.artifacts_dir, self.tsung_controller_id)
        if not os.path.exists(self.tsung_artifacts_basedir):
            os.makedirs(self.tsung_artifacts_basedir)
        else:
            msg = "Tsung artifacts basedir already exists, will not create: %s"
            self.log.warning(msg, self.tsung_artifacts_basedir)

        self.reader = TsungStatsReader(self.tsung_artifacts_basedir, self.log)
        if isinstance(self.engine.aggregator, ConsolidatingAggregator):
            self.engine.aggregator.add_underling(self.reader)

        self.stdout = open(self.engine.create_artifact("tsung", ".out"), 'w')
        self.stderr = open(self.engine.create_artifact("tsung", ".err"), 'w')

    def __modify_user_tsung_config(self, user_config_path):
        modified_config_path = self.engine.create_artifact("tsung-config", ".xml")
        load = self.get_load()
        config = TsungConfig(self.tool)
        config.load(user_config_path)
        config.apply_load_profile(load)
        config.apply_dumpstats()
        config.save(modified_config_path)
        return modified_config_path

    def _generate_tsung_config(self):
        config_file = self.engine.create_artifact("tsung-config", ".xml")
        scenario = self.get_scenario()
        load = self.get_load()
        config = TsungConfig(self.tool)
        config.generate(scenario, load)
        config.save(config_file)
        return config_file

    def startup(self):
        args = [
            self.tool.tool_path,
            '-f', self.tsung_config,
            '-l', self.tsung_artifacts_basedir,
            '-i', self.tsung_controller_id,
            '-w', '0',
        ]

        user_cmd = self.settings.get("cmdline")
        if user_cmd:
            args += user_cmd.split(" ")

        args.append('start')

        self.process = self._execute(args)

    def check(self):
        ret_code = self.process.poll()
        if ret_code is None:
            return False
        if ret_code != 0:
            raise ToolError("Tsung exited with non-zero code: %s" % ret_code, self.get_error_diagnostics())
        return True

    def shutdown(self):
        shutdown_process(self.process, self.log)

    def install_required_tools(self):
        self.tool = self._get_tool(Tsung, config=self.settings)
        if not self.tool.check_if_installed():
            self.tool.install()

    def get_widget(self):
        if not self.widget:
            self.widget = ExecutorWidget(self, self.tsung_controller_id)
        return self.widget

    def resource_files(self):
        script = self.get_script_path()
        if script:
            return [script]
        else:
            return []

    def get_error_diagnostics(self):
        diagnostics = []
        if self.stdout is not None:
            with open(self.stdout.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Tsung STDOUT:\n" + contents)
        if self.stderr is not None:
            with open(self.stderr.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Tsung STDERR:\n" + contents)
        return diagnostics


class TsungStatsReader(ResultsReader):
    def __init__(self, tsung_basedir, parent_logger):
        super(TsungStatsReader, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.tsung_basedir = tsung_basedir
        self.stats_file = FileReader(parent_logger=self.log, file_opener=self.open_stats)
        self.log_file = FileReader(parent_logger=self.log, file_opener=self.open_log)
        self.delimiter = ";"
        self.partial_buffer = ""
        self.skipped_header = False
        self.concurrency = 0

    def open_stats(self, filename):
        return self.open_file(ext='dump')

    def open_log(self, filename):
        return self.open_file(ext='log')

    def open_file(self, ext):
        basedir_contents = os.listdir(self.tsung_basedir)

        if not basedir_contents:
            self.log.debug("Tsung artifacts not appeared yet")
            return

        if len(basedir_contents) != 1:
            self.log.warning("Multiple files in Tsung basedir %s, this shouldn't happen", self.tsung_basedir)
            return

        filename = os.path.join(self.tsung_basedir, basedir_contents[0], "tsung." + ext)

        if not os.path.isfile(filename):
            self.log.debug("File not appeared yet: %s", filename)
            return
        if not os.path.getsize(filename):
            self.log.debug("File is empty: %s", filename)
            return

        self.log.debug('Opening file: %s', filename)
        return open(filename, mode='rb')

    def _read_concurrency(self, last_pass):
        lines = self.log_file.get_lines(size=1024 * 1024, last_pass=last_pass)
        extractor = re.compile(r'^stats: users (\d+) (\d+)$')

        for line in lines:
            match = extractor.match(line.strip())
            if not match:
                continue
            self.concurrency = int(match.group(2))
            self.log.debug("Actual Tsung concurrency: %s", self.concurrency)

    def _read(self, last_pass=False):
        self.log.debug("Reading Tsung results")

        self._read_concurrency(last_pass)
        lines = self.stats_file.get_lines(size=1024 * 1024, last_pass=last_pass)

        for line in lines:
            if not line.endswith("\n"):
                self.partial_buffer += line
                continue

            if not self.skipped_header and line.startswith("#"):
                self.skipped_header = True
                continue

            line = "%s%s" % (self.partial_buffer, line)
            self.partial_buffer = ""

            line = line.strip()
            fields = line.split(self.delimiter)

            tstamp = int(float(fields[0]))
            url = fields[4] + fields[5]
            rstatus = fields[6]
            rsize = int(fields[7])
            etime = float(fields[8]) / 1000.0
            trname = fields[9]
            error = fields[10] or None

            con_time = 0
            latency = 0

            yield tstamp, url, self.concurrency, etime, con_time, latency, rstatus, error, trname, rsize


class TsungConfig(object):
    def __init__(self, tsung_tool):
        self.log = logging.getLogger(self.__class__.__name__)
        self.root = etree.Element("tsung", loglevel="notice", version="1.0", dumptraffic="protocol", backend="text")
        self.tree = etree.ElementTree(self.root)
        self.tool = tsung_tool

    def load(self, filename):
        try:
            self.tree = etree.ElementTree()
            self.tree.parse(filename)
            self.root = self.tree.getroot()
        except BaseException as exc:
            self.log.debug("Tsung: XML parsing error: %s", traceback.format_exc())
            raise TaurusInternalException("Tsung: XML parsing failed for file %s: %s" % (filename, exc))

    def save(self, filename):
        self.log.debug("Saving Tsung config to: %s", filename)
        with open(filename, "wb") as fhd:
            tsung_dtd = self.tool.get_dtd_path()
            self.log.debug("Detected Tsung DTD: %s", tsung_dtd)
            doctype = '<!DOCTYPE tsung SYSTEM "%s">' % tsung_dtd
            xml = etree.tostring(self.tree, pretty_print=True, encoding="UTF-8", xml_declaration=True, doctype=doctype)
            fhd.write(xml)

    def generate(self, scenario, load):
        self.root.append(self.__gen_clients())
        self.root.append(self.__gen_servers(scenario))
        self.root.append(self.__gen_load(load))
        self.root.append(self.__gen_options(scenario))
        self.root.append(self.__gen_sessions(scenario))

    def apply_dumpstats(self):
        self.root.set("dumptraffic", "protocol")
        self.root.set("backend", "text")

    def apply_load_profile(self, load):
        # do not apply unspecified load profile
        if not load.concurrency and not load.hold:
            return
        original_load = self.find("//tsung/load")
        generated_load = self.__gen_load(load)
        if not original_load:
            self.log.warning("<load> section not found in Tsung config, will create one")
            servers = self.find("//tsung/servers")
            if not servers:
                raise TaurusConfigError("Provided Tsung script is invalid: <servers> section not found")
            servers.addnext(generated_load)
        else:
            self.root.replace(original_load[0], generated_load)

    @staticmethod
    def __time_to_tsung_time(time_amount):
        if time_amount % 3600 == 0:
            return time_amount // 3600, "hour"
        elif time_amount % 60 == 0:
            return time_amount // 60, "minute"
        else:
            return time_amount, "second"

    def __gen_clients(self):
        # TODO: distributed clients?
        clients = etree.Element("clients")
        client = etree.Element("client", host="localhost", use_controller_vm="true")
        clients.append(client)
        return clients

    def __first_http_request(self, scenario):
        for request in scenario.get_requests():
            if isinstance(request, HTTPRequest):
                return request
        return None

    def __gen_servers(self, scenario):
        default_address = scenario.get("default-address")
        if default_address:
            base_addr = parse.urlparse(default_address)
        else:
            first_request = self.__first_http_request(scenario)
            if not first_request:
                raise TaurusConfigError("Tsung: you must specify requests in scenario")
            base_addr = parse.urlparse(first_request.url)
            self.log.debug("default-address was not specified, using %s instead", base_addr.hostname)

        servers = etree.Element("servers")
        port = base_addr.port if base_addr.port is not None else 80
        server = etree.Element("server", host=base_addr.hostname, port=str(port), type="tcp")
        servers.append(server)
        return servers

    def __gen_load(self, load):
        """
        Generate Tsung load profile.

        Tsung load progression is scenario-based. Virtual users are erlang processes which are spawned according to
        load profile. Each user executes assigned session (requests + think-time + logic) and then dies.
        :param scenario:
        :param load:
        :return:
        """
        concurrency = load.concurrency if load.concurrency is not None else 1
        load_elem = etree.Element("load")
        if load.duration:
            duration, unit = self.__time_to_tsung_time(int(round(load.duration)))
            load_elem.set('duration', str(duration))
            load_elem.set('unit', unit)
        phases = []

        if load.hold:
            duration, unit = self.__time_to_tsung_time(int(round(load.hold)))
            users = etree.Element("users", arrivalrate=str(concurrency), unit="second")
            phase = etree.Element("arrivalphase",
                                  phase=str("1"),
                                  duration=str(duration),
                                  unit=unit)
            phase.append(users)
            phases.append(phase)
        else:
            raise TaurusConfigError("Tsung: you must specify test duration with 'hold-for'")

        for phase in phases:
            load_elem.append(phase)

        return load_elem

    def __gen_options(self, scenario):
        options = etree.Element("options")

        global_think_time = scenario.get_think_time()
        if global_think_time:
            think_time = int(dehumanize_time(global_think_time))
            options.append(etree.Element("option", name="thinktime", value=str(think_time), random="false"))

        global_tcp_timeout = scenario.get("timeout")
        if global_tcp_timeout:
            timeout = int(dehumanize_time(global_tcp_timeout) * 1000)
            options.append(etree.Element("option", name="connect_timeout", value=str(timeout)))

        global_max_retries = scenario.get("max-retries", 1)
        options.append(etree.Element("option", name="max_retries", value=str(global_max_retries)))
        return options

    def __gen_sessions(self, scenario):
        sessions = etree.Element("sessions")
        session = etree.Element("session", name="taurus_requests", probability="100", type="ts_http")
        for request in scenario.get_requests():
            if not isinstance(request, HTTPRequest):
                msg = "Tsung config generator doesn't support '%s' blocks, skipping"
                self.log.warning(msg, request.NAME)
                continue

            request_elem = etree.Element("request")
            http_elem = etree.Element("http", url=request.url, method=request.method, version="1.1")
            if request.body:
                http_elem.set('contents', request.body)

            headers = copy.deepcopy(scenario.get_headers())
            headers.update(copy.deepcopy(request.headers))
            for header_name, header_value in iteritems(headers):
                http_elem.append(etree.Element("http_header", name=header_name, value=header_value))

            request_elem.append(http_elem)
            session.append(request_elem)
            if request.get_think_time():
                think_time = int(dehumanize_time(request.get_think_time()))
                session.append(etree.Element("thinktime", value=str(think_time), random="false"))
        sessions.append(session)
        return sessions

    def find(self, xpath_selector):
        return self.tree.xpath(xpath_selector)


class Tsung(RequiredTool):
    INSTALLATION_DOCS = "http://gettaurus.org/docs/Tsung/#Tsung-Installation"
    DEFAULT_DTD_PATH = "/usr/share/tsung/tsung-1.0.dtd"

    def __init__(self, config=None, **kwargs):
        settings = config or {}
        tool_path = settings.get("path", "tsung")
        super(Tsung, self).__init__(tool_path=tool_path, installable=False, **kwargs)

    def check_if_installed(self):
        self.log.debug("Trying %s...", self.tool_name)
        try:
            out, err = self.call([self.tool_path, "-v"])
        except CALL_PROBLEMS as exc:
            self.log.warning("%s check failed: %s", self.tool_name, exc)
            self.log.warning("Info for tsung installation: %s", self.INSTALLATION_DOCS)
            return False

        if err:
            out += err
        self.log.debug("%s output: %s", self.tool_name, out)
        return True

    def get_tool_abspath(self):
        if not self.tool_path:
            return None

        abspath = os.path.abspath(self.tool_path)
        if os.path.exists(abspath):
            return abspath

        return which(self.tool_path)


    @staticmethod
    def get_tool_prefix(tool_abspath):
        """
        Get tsung installation prefix (like /usr/ or /usr/local)
        :return: str
        """
        if tool_abspath is None:
            return None

        parts = tool_abspath.split(os.sep)

        if len(parts) < 2:
            return None

        # cut 'bin/tsung' from abspath
        prefix = os.sep.join(parts[:-2])
        return prefix

    def get_dtd_path(self):
        "Get path of DTD validation file for Tsung."
        tsung_abspath = self.get_tool_abspath()
        prefix = self.get_tool_prefix(tsung_abspath)
        if not prefix:
            return self.DEFAULT_DTD_PATH
        else:
            return os.path.join(prefix, "share", "tsung", "tsung-1.0.dtd")
